import io
from dataclasses import dataclass
from datetime import datetime, timezone
from pickle import dump, load

import pandas as pd

from trafficdrop.detector.common_types import TrafficDirection
from trafficdrop.detector.const import Const
from trafficdrop.detector.utils import DfLogger, Direction, Time


@dataclass
class TrafficDropBaseline:
    """
    This is the Traffic drop baseline object that is persisted in the minio.
    For Traffic drop, we have 2 unique DF representing src & dst vm.

    We will call pickle.dump(...) to turn that into a io.ByteIO
    """
    src_baseline: io.BytesIO  # pandas dataframe in parquet format
    dst_baseline: io.BytesIO  # pandas dataframe in parquet format


class Baseline:
    """
    TODO: consider moving these functions to a common utils
    """

    @staticmethod
    def create_traffic_drop_baseline_buffer(src_df: pd.DataFrame, dst_df: pd.DataFrame) -> io.BytesIO | None:
        src_buffer = io.BytesIO()
        src_df.to_parquet(src_buffer)
        src_buffer.seek(0)  # rewind the buffer

        dst_buffer = io.BytesIO()
        dst_df.to_parquet(dst_buffer)
        dst_buffer.seek(0)  # rewind the buffer

        merged_buffer = io.BytesIO()
        merged_traffic_drop_baseline = TrafficDropBaseline(
            src_buffer, dst_buffer)

        dump(merged_traffic_drop_baseline, merged_buffer)
        merged_buffer.seek(0)  # rewind the buffer

        return merged_buffer

    @staticmethod
    def transform_combined_baseline(merged_baselines: pd.DataFrame, site_id: str,
                                    direction: TrafficDirection) -> pd.DataFrame:
        """
        convert merged baseline (df from druid) into a 'derived' df containing summary
        stats:
        - vm, total_blocked std, total_blocked mean, blockedRatio Mean. ...., allowMean
        :param site_id:
        :param merged_baselines: druid query df
        :param direction: src | dst query
        :return: a df containing summary statistic of the druid query
        """
        derived_results = pd.DataFrame()
        group_by_column = Direction.get_direction_vm_dimension(direction)

        # for single result, we mark the std as zero
        derived_results[Const.BLOCKED_STD] = \
            merged_baselines.groupby(group_by_column)[
            Const.TOTAL_BLOCKED].std().fillna(0)
        derived_results[Const.BLOCKED_MEAN] = \
            merged_baselines.groupby(group_by_column)[
            Const.TOTAL_BLOCKED].mean()
        derived_results[Const.BLOCKED_RATIO_MEAN] = \
            merged_baselines.groupby(group_by_column)[Const.BLOCK_RATIO].mean()
        derived_results[Const.ALLOWED_RATIO_MEAN] = \
            merged_baselines.groupby(group_by_column)[Const.ALLOW_RATIO].mean()
        derived_results[Const.ALLOWED_MEAN] = \
            merged_baselines.groupby(group_by_column)[
            Const.TOTAL_ALLOWED].mean()
        # TODO: confirm the usefulness of this
        #       return the # days since the earliest record : (now - oldest_record).days
        derived_results[Const.BASELINE_DURATION] = (
            merged_baselines
            .groupby(group_by_column)[Const.TIMESTAMP]
            .min()
            .apply(lambda time_str: (datetime.now(timezone.utc) - Time.parse_iso_datetime(time_str)).days))

        """
        Derived DF:

        | vm    | blockedStdDeviation | blockedMean | blockedRatioMean | allowRatioMean | allowMean | baselineDuration |
        |:------|--------------------:|------------:|-----------------:|---------------:|----------:|-----------------:|
        | vm-0  |                0    |       0     |         0        |       1        |  14       |                4 |
        | vm-1  |             1036.45 |     271.467 |         0.371075 |       0.628925 |  10.2667  |               29 |
        | vm-2  |                0    |       0     |         0        |       1        |   2       |                5 |
        | vm-3  |             1033.09 |     280.207 |         0.390633 |       0.609367 |   8.65517 |               30 |
        """
        DfLogger.log(derived_results, site_id, direction, "transform-derived")

        return derived_results

    @staticmethod
    def get_combined_df_from(baseline: io.BytesIO) -> tuple[pd.DataFrame, pd.DataFrame]:
        traffic_drop_baseline = load(baseline)
        src_baseline = pd.read_parquet(traffic_drop_baseline.src_baseline)
        dst_baseline = pd.read_parquet(traffic_drop_baseline.dst_baseline)

        return src_baseline, dst_baseline
