import io
import math
from datetime import date, datetime, timedelta
from pickle import dump, load
from typing import Any, Generator

import numpy as np
import pandas as pd

from common.utils import logging
from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import Interval, PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.detector.core import BatchDetector, DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import EventInfo
from common.utils.flow.core import FlowService
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    DetectorRunStatus,
    DetectorStatusLabel,
    ExclusionLabel,
    ExclusionTypeInfo,
    UserConfigInfo,
)
from trafficdrop.detector.baseline import Baseline, TrafficDropBaseline
from trafficdrop.detector.common_types import TrafficDirection
from trafficdrop.detector.config import Config
from trafficdrop.detector.const import Const
from trafficdrop.detector.event import Builder
from trafficdrop.detector.query import DruidQuery
from trafficdrop.detector.utils import DfLogger, Direction


class TrafficDrop(BatchDetector):
    """
    Main Detector Class (Implements BatchDetector)

    It look at INCOMING (DST) & OUTGOING (SRC) of a flow and calculate statistically values
    associated with ite drop count

    The current implementation combined a weighted score from
      1) Std from block count
      2) block trend
      3) port cardinality
    To see determined if there is significant drop traffic changes
    """

    logger = logging.get_common_logger(__name__)

    def __init__(
        self,
        baseline_service: BaselineService,
        checkpoint_service: CheckpointService,
        user_config_service: UserConfigService,
        event_service: EventService,
        flow_service: FlowService,
        lookup_service: LookupService,
        now: datetime,
    ):
        super().__init__(baseline_service, checkpoint_service, user_config_service, event_service, now)
        self._flow_service = flow_service
        self._lookup_service = lookup_service
        self._now = now
        self._beginning_of_today = datetime.combine(now.date(), datetime.min.time())
        self._query = DruidQuery(self.detector_name().name, self._flow_service)
        self._total_events_published = 0

    @staticmethod
    def detector_name() -> DetectorType:
        return DetectorType.NETWORK_TRAFFIC_DROP

    @staticmethod
    def _baseline_max_retention_time() -> timedelta:
        return timedelta(days=float(Config.BASELINE_MAX_RETENTION_DAYS))

    def _generate_baseline(self, site_id: str, interval: Interval) -> PartialBaseline:
        TrafficDrop.logger.info(f"generating traffic drop baselines for interval {interval}")

        # Run src query
        src_query_result = self._execute_traffic_drop_query([interval], site_id, TrafficDirection.SRC)
        TrafficDrop.logger.info(
            f"got {len(src_query_result.index)} rows from druid when building baseline for SRC({interval})"
        )
        src_buffer = io.BytesIO()
        src_query_result.to_parquet(src_buffer)

        # Run dst query
        dst_query_result = self._execute_traffic_drop_query([interval], site_id, TrafficDirection.DST)
        TrafficDrop.logger.info(
            f"got {len(dst_query_result.index)} rows from druid when building baseline for DST({interval})"
        )
        dst_buffer = io.BytesIO()
        dst_query_result.to_parquet(dst_buffer)

        # combine query results
        buffer = io.BytesIO()
        final_baseline = TrafficDropBaseline(src_buffer, dst_buffer)
        dump(final_baseline, buffer)
        return PartialBaseline(interval, buffer)

    def _get_baseline_intervals(self, missing_intervals: list[Interval]) -> list[Interval]:
        return BaselineService.convert_to_daily_intervals_with_trim(
            missing_intervals,
            self._beginning_of_today,
            timedelta(days=float(Config.BASELINE_MAX_HISTORICAL_CREATION_DAYS)),
        )

    def _execute_traffic_drop_query(
        self,
        intervals: list[Interval],
        site_id: str,
        direction: TrafficDirection,
        included_vms: list[str] = [],
        excluded_vms: list[str] = [],
    ) -> pd.DataFrame:
        """
        Query DF

        | srcVmId (idx) | dstFwAllowed | extDstAllowed | combinedDstAllowed | allowRatio | totalBlocked | srcFwAllowed |
        | totalAllowed | srcFwBlocked | combinedSrcAllowed | dstFwBlocked | blockRatio | total | extSrcAllowed
        | timestamp |
        """
        return self._query.execute_base_traffic_drop_query(intervals, site_id, direction, included_vms, excluded_vms)

    def _combine_baseline_objects(self, baselines: list[PartialBaseline]) -> io.BytesIO | None:
        combined_baseline = pd.DataFrame()
        combined_baseline_buffer = io.BytesIO()

        if len(baselines) == 0:
            TrafficDrop.logger.warn("got empty list of baselines to combine, returning empty baseline for detection")
            combined_baseline.to_parquet(combined_baseline_buffer)
            return combined_baseline_buffer

        src_merged_baselines = pd.DataFrame()
        dst_merged_baselines = pd.DataFrame()
        for partial_baseline in baselines:
            partial_traffic_drop_baseline = load(partial_baseline.data)

            src_partial_baseline = pd.read_parquet(partial_traffic_drop_baseline.src_baseline)
            dst_partial_baseline = pd.read_parquet(partial_traffic_drop_baseline.dst_baseline)

            # append interval baseline
            src_merged_baselines = pd.concat([src_merged_baselines, src_partial_baseline])
            dst_merged_baselines = pd.concat([dst_merged_baselines, dst_partial_baseline])

        # TODO: change _combine_baseline_objects api signature to pass in the site_id, instead of passing emtpy
        DfLogger.log(src_merged_baselines, "", TrafficDirection.SRC, "merged-baseline")
        DfLogger.log(dst_merged_baselines, "", TrafficDirection.DST, "merged-baseline")

        """ Combined Baseline: (index srcVmId/dstVmId)

        | src/dstVmId | dstFwAllowed | extDstAllowed | combinedDstAllowed | allowRatio | totalBlocked | srcFwAllowed |
        |:------------|-------------:|--------------:|-------------------:|-----------:|-------------:|-------------:|
        | vm-0        |            6 |             0 |                  6 | 0.00298805 |         4004 |            6 |

        | totalAllowed | srcFwBlocked | combinedSrcAllowed | dstFwBlocked | blockRatio | total | extSrcAllowed |
        |-------------:|-------------:|-------------------:|-------------:|-----------:|------:|--------------:|
        |           12 |         4004 |                  6 |            0 |  0.997012  |  4016 |             0 |

        | timestamp                |
        |:-------------------------|
        | 2023-04-23T00:00:00.000Z |
        """

        # return baseline into byteIO
        return Baseline.create_traffic_drop_baseline_buffer(src_merged_baselines, dst_merged_baselines)

    def _get_day_start_to_now_interval(self) -> Interval:
        return Interval(self._beginning_of_today, self._now)

    def _get_baseline_start_datetime(self) -> datetime:
        delta = self._baseline_max_retention_time()
        return self._beginning_of_today - delta

    def _get_baseline_to_now_interval(self) -> Interval:
        baseline_start = self._get_baseline_start_datetime()
        return Interval(baseline_start, self._now)

    def _get_baseline_interval(self) -> Interval:
        # return day granularity: e.g. (30days ago, start of day today)
        delta = self._baseline_max_retention_time()
        today_midnight = self._beginning_of_today

        return Interval(today_midnight - delta, today_midnight)

    def _get_epoch_to_baseline_start_interval(self) -> Interval:
        baseline_start = self._get_baseline_start_datetime()
        return Interval(datetime.utcfromtimestamp(0), baseline_start)

    def _get_excluded_computes_from_config(self, config: UserConfigInfo) -> list[str]:
        if config.exclusion_settings == {}:
            return []

        excluded_computes = []
        for setting in config.exclusion_settings.exclusions:
            # technically, this can be ExclusionLabel.DESTINATION or SOURCE but the existing API
            # is using NONE (traffic drop looks at both directions of traffic).
            # In order to avoid changing the API contract, we allow
            # DESTINATION, SOURCE or NONE (existing behavior)
            if (
                setting.label == ExclusionLabel.DESTINATION
                or setting.label == ExclusionLabel.SOURCE
                or setting.label == ExclusionLabel.NONE
            ):
                # setting.excluded_items_info.value can technically be any type,
                # but here we're assuming it's a list of str
                excluded_entity_ids: list[str] = setting.excluded_items_info.value  # type: ignore (suppress linter warning)
                if setting.type == ExclusionTypeInfo.VM:
                    excluded_computes += excluded_entity_ids
                elif setting.type == ExclusionTypeInfo.GROUP:
                    excluded_computes += self._lookup_service.groups_to_computes_lookup(
                        config.site_id, excluded_entity_ids
                    )
        return excluded_computes

    def _filter_today_result(
        self,
        site_id: str,
        user_config: UserConfigInfo,
        direction: TrafficDirection,
        baseline_df: pd.DataFrame,
        today_df: pd.DataFrame,
    ) -> pd.DataFrame:
        """
        1) Look at 30 day baseline & identify compute that has less than 3 days worth
           of baseline. Then we remove those from today results
        2) excluded vm based on settings

        :param direction:
        :param baseline_df:
        :param today_df:
        :return: a new today df, filtering out VM that has less 3 days of baseline
        """
        compute_dimension = Direction.get_direction_vm_dimension(direction)

        # df[vm,oldest_timestamp(in date)]
        compute_oldest_df = (
            baseline_df.groupby(compute_dimension)[Const.TIMESTAMP]
            .min()
            .transform(lambda x: pd.to_datetime(x).dt.date)
            .to_frame()
        )

        DfLogger.log(compute_oldest_df, site_id, direction, "oldest-compute")

        # find out all compute with base older than TIME_DELTA_DAYS (we want to keep VM with enough baseline)
        compute_with_enough_baselines = compute_oldest_df.loc[
            compute_oldest_df[Const.TIMESTAMP] < (date.today() - timedelta(days=int(Config.MIN_BASELINE_DURATION_DAYS)))
        ]

        DfLogger.log(compute_with_enough_baselines, site_id, direction, "compute-enough-baseline")

        # keep compute that have enough baseline data
        today_df = today_df[today_df.index.isin(compute_with_enough_baselines.index)]

        # process exclusion (drop vm based on vm exclusion settings)
        excluded_computes = self._get_excluded_computes_from_config(user_config)
        today_df = today_df[~today_df.index.isin(excluded_computes)]

        DfLogger.log(today_df, site_id, direction, "filtered-today")

        return today_df

    def _detect_events(
        self, user_config: UserConfigInfo, baseline: io.BytesIO, checkpoint: io.BytesIO | None
    ) -> Generator[EventInfo, Any, Any]:
        """
        The main entry point for traffic drop detection
        1. Get the past 30 day drop statistic, transform to a 'derived' df (with std etc)
        2. Get today drop statistics
        3. Compare today & derived statistic (see _detect_traffic_drop_events)

        :param user_config:
        :param baseline:
        :return:
        """
        TrafficDrop.logger.info(f"Running Traffic Drop detection for {user_config.site_id}")

        # convert traffic drop baseline back into 2 src & dst df
        src_baseline, dst_baseline = Baseline.get_combined_df_from(baseline)

        site_id = user_config.site_id

        # Calculate today data
        midnight_to_now = self._get_day_start_to_now_interval()

        src_today_query_result = self._execute_traffic_drop_query([midnight_to_now], site_id, TrafficDirection.SRC)

        dst_today_query_result = self._execute_traffic_drop_query([midnight_to_now], site_id, TrafficDirection.DST)

        # filter out VM that are not eligible (which doesn't contain enough baseline data)
        src_today_query_result = self._filter_today_result(
            site_id, user_config, TrafficDirection.SRC, src_baseline, src_today_query_result
        )
        dst_today_query_result = self._filter_today_result(
            site_id, user_config, TrafficDirection.DST, dst_baseline, dst_today_query_result
        )

        # TODO: check today result is unique per VM, drop duplicates if needed
        #       it shouldn't happen since the today result is group by Day granularity
        DfLogger.log(src_today_query_result, site_id, TrafficDirection.SRC, "final-today")
        DfLogger.log(dst_today_query_result, site_id, TrafficDirection.DST, "final-today")

        # generate derived statistics from merged baseline
        src_derived_baseline = Baseline.transform_combined_baseline(src_baseline, site_id, TrafficDirection.SRC)
        dst_derived_baseline = Baseline.transform_combined_baseline(dst_baseline, site_id, TrafficDirection.DST)

        DfLogger.log(src_derived_baseline, site_id, TrafficDirection.SRC, "derived-baseline")
        DfLogger.log(dst_derived_baseline, site_id, TrafficDirection.DST, "derived-baseline")

        # run src & dst detection by comparing today & derived df)
        for event in self._detect_traffic_drop_events(
            user_config, TrafficDirection.SRC, src_derived_baseline, src_today_query_result
        ):
            yield event

        for event in self._detect_traffic_drop_events(
            user_config, TrafficDirection.DST, dst_derived_baseline, dst_today_query_result
        ):
            yield event

    def _detect_traffic_drop_events(
        self, user_config: UserConfigInfo, direction: TrafficDirection, derived_df: pd.DataFrame, today_df: pd.DataFrame
    ) -> Generator[EventInfo, Any, Any]:
        """
        Compare derived and today statistics and generate a score via 3 stages
        - Sigma & StdDev (between today & 30 day derived statistics)
        - Trend
        - Port Cardinality
        Finally combine all result & generate a score (based on fixed weightage)

        :param user_config:
        :param direction:
        :param derived_df: 30 days baseline 'summary' statistics
        :param today_df: today traffic drop
        :return:
        """
        # Return quickly if there is no blocked flow today is empty
        if len(today_df.index) == 0:
            TrafficDrop.logger.info("Discontinuing this job since no blocked flows are found today")
            return

        # executeBlockedCountStdDeviationStage:
        #
        # Step 1: Check sigma std deviation for blocked count
        """
        | src/dst VmId |   sigmaStdDev |   stdDevScore |
        |:-------------|--------------:|--------------:|
        | vm-0         |       1.67643 |       37.3768 |
        """
        stage1_score = self._execute_blocked_count_std_stage(derived_df, today_df, user_config.site_id, direction)

        # executeTrendChangeStage
        #
        # Step 2 : Check if the block/total trend has flipped. This stage helps with the case where some previously
        #          allowed traffic are now getting blocked
        """
        | src/dst VmId |   trendChange |   trendChangeDiff |   trendChangeNormFactor |   totalMean |
        |:-------------|--------------:|------------------:|------------------------:|------------:|
        | vm-0         |           100 |           712.039 |                  7.1486 |     281.733 |
        """
        stage2_score = self._execute_trend_change_stage(derived_df, today_df, user_config.site_id, direction)

        # executePortCardinalityStage
        #
        # Step 3 : Check if all the drops today are caused by a new port/protocol/app
        #          Check if all the drops today are caused by a new port/protocol/app
        #          In this step we will only check for those VMs that are suspects from Step 1 and 2.
        #          Compare cardinality of blocked ports in the last 30 days with cardinality of blocked counts today.
        """
        | src/dst VmId |   cardinalityCount | timestamp                |
        |:-------------|-------------------:|:-------------------------|
        | vm-0         |             3.0022 | 2023-04-23T00:00:00.000Z |
        """
        stage3_score = self._execute_port_cardinality_stage(stage2_score, user_config.site_id, direction)

        # calculate score based on 3 stages
        """
        | src/dst VmId |   stdDevScore |   trendChange |   cardinalityScore |   Score | is_silent   |
        |:-------------|--------------:|--------------:|-------------------:|--------:|:------------|
        | vm-0         |       37.3768 |           100 |                  0 | 43.6884 | True        |
        """
        final_score = self._combine_score(
            stage1_score, stage2_score, stage3_score, user_config, user_config.site_id, direction
        )

        DfLogger.log(final_score, user_config.site_id, direction, "final score")

        # generate event
        if len(final_score.index) == 0:
            TrafficDrop.logger.warn("combined score is empty. No event will be generated")
            return

        tag = f"{user_config.site_id}:{direction}"
        DfLogger.log(final_score, user_config.site_id, direction, "TrafficDrop Event")

        # prepare a compute->ip lookup table for all events
        compute_to_ip_lookup = self._lookup_service.computes_to_ip_lookup(
            user_config.site_id, self._get_day_start_to_now_interval()
        )
        TrafficDrop.logger.info(f"preparing compute to ip lookup (size:{len(compute_to_ip_lookup)})")

        for compute_id, row in final_score.iterrows():
            self._total_events_published += 1
            if self._total_events_published % 100 == 0:
                TrafficDrop.logger.info(f"Number of events published so far: {self._total_events_published}")
            if self._total_events_published >= int(Config.NUMBER_OF_PERMITTED_EVENTS):
                TrafficDrop.logger.info(f"{Config.NUMBER_OF_PERMITTED_EVENTS} events reached hence stopping publishing")
                break
            TrafficDrop.logger.info(f"creating {self.detector_name()} event for compute: {compute_id}({tag})")
            yield self._create_event(str(compute_id), row, user_config.site_id, direction, compute_to_ip_lookup)
        TrafficDrop.logger.info(f"Number of events published so finally: {self._total_events_published}")

    @staticmethod
    def sigma_under_12(sigma):
        # sigmaUnder12 = (double sigma) -> Math.min(Math.sqrt(STDDEV_QUADRATIC_COEFFICIENT_LESSEQ_12*sigma), 100);
        return np.minimum(math.sqrt(Const.STDDEV_QUADRATIC_COEFFICIENT_LESSEQ_12 * sigma), 100)

    @staticmethod
    def sigma_above_12(sigma):
        # sigmaAbove12 = (double sigma) -> Math.min((Math.pow(sigma, 2) - 24*sigma +15744)/156, 150);
        return np.minimum((math.pow(sigma, 2) - 24 * sigma + 15744) / 156, 150)

    def _execute_blocked_count_std_stage(
        self, derived_df: pd.DataFrame, today_df: pd.DataFrame, site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        tag = f"{site_id}:{direction}"
        TrafficDrop.logger.info(f"STEP 1 ({tag}) : Check if blocked count is at least 3 SD")
        """
        Compare today blocked vs baseline mean blocked count
          if today blocked > baseline mean
              Find out how much it deviates from mean (sigma)
              The deviation ratio R: (total_blocked - blocked_mean)/max(5, blocked_std))
          sigma > 12 vs < 12 changes the score
          add score weighted score
        """

        compute_id = Direction.get_direction_vm_dimension(direction)

        # join today with derived stats (each row represent 1 vm)
        result = pd.merge(today_df, derived_df, on=compute_id)
        if len(result.index) == 0:
            TrafficDrop.logger.warn("stage1: no compute id overlaps between today & baseline")
            empty_df = pd.DataFrame(columns=[compute_id, Const.BLOCKED_STD_SCORE])
            empty_df.set_index(compute_id, inplace=True)
            return empty_df

        # filter rows that have +ve std dev (today block > baseline mean)
        result.query(f"{Const.TOTAL_BLOCKED} > {Const.BLOCKED_MEAN}", inplace=True)

        if len(result.index) == 0:
            TrafficDrop.logger.warn(f"{Const.TOTAL_BLOCKED} > {Const.BLOCKED_MEAN}," f" returning empty df for stage1")
            result[Const.BLOCKED_STD_SCORE] = None
            return result  # empty df

        # sigmaStdDev : (deviation-from-mean)/max(MIN_STD, blocked_std)
        result[Const.SIGMA_STD_DEV] = result.apply(
            lambda d: (d[Const.TOTAL_BLOCKED] - d[Const.BLOCKED_MEAN])
            / np.maximum(Const.MIN_STD_DEVIATION, d[Const.BLOCKED_STD]),
            axis=1,
        )

        # stdDevScore = below12(sigmaStdDev) or above12(sigmaStdDev)
        result[Const.BLOCKED_STD_SCORE] = result.apply(
            lambda d: TrafficDrop.sigma_under_12(d[Const.SIGMA_STD_DEV])
            if (d[Const.SIGMA_STD_DEV] < 12)
            else TrafficDrop.sigma_above_12(d[Const.SIGMA_STD_DEV]),
            axis=1,
        )

        DfLogger.log(result[[Const.SIGMA_STD_DEV, Const.BLOCKED_STD_SCORE]], site_id, direction, "stage1-score")

        # return vm, score only, assign index
        return result[[Const.BLOCKED_STD_SCORE]]

    def _execute_trend_change_stage(
        self, derived_df: pd.DataFrame, today_df: pd.DataFrame, site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        tag = f"{site_id}:{direction}"
        TrafficDrop.logger.info(f"STEP 2 ({tag}): Check trend change in allowed/blocked flows")
        """
        For each VM
          B: get baseline allowRatio Mean
          T: today sum of allowed_ratio
          if T < B (allow drop compares to baseline)
            totol_mean = BL blocked mean + allowed mean
            normalized factor = 1 or today total block + today total allowed/totol_mean
            <some annoy math>

            score = diffPercent
            add score weighted score
        """

        vm_id = Direction.get_direction_vm_dimension(direction)

        # join today with derived stats (each row represent 1 vm)
        result = pd.merge(today_df, derived_df, on=vm_id)

        # filter out If today's allowed ratio is less than mean allowed ratio
        result.query(f"{Const.ALLOW_RATIO} < {Const.ALLOWED_RATIO_MEAN}", inplace=True)

        if len(result.index) == 0:
            TrafficDrop.logger.debug(
                f"{Const.ALLOW_RATIO} < {Const.ALLOWED_RATIO_MEAN}," f" returning empty df for stage2"
            )
            result[Const.TREND_CHANGE_SCORE] = None
            return result  # empty df

        """
        TODO :
        There is one case where this score can result in false positives. The ratio may intermittently flip during
        the day, because all the traffic for that day is not seen yet. In these cases,
        we need to dilute the effect of this step because the ratios could stabilize later on in the day.
        Eg:
        Baseline Mean Allowed = 30, Blocked = 20, Mean AllowRatio=0.6
        Today at 2 PM , Allowed = 2, Blocked = 1, AllowRatio=0.28
        Here it appears as if today's allow ratio is drastically lesser than baseline allow ratio and the diff % is 53%.
        However, the total flows today is only 3. We need to wait until the rest of the day is over to see if allowRatio
        is indeed lesser today. Adding a normalizing factor =(today's total / mean total) will help dilute the effect
        until an expected total flow count is reached.
        """

        # factor : (today blocked + allowed) / (baseline blocked + allowed)
        result[Const.TOTAL_MEAN] = result[Const.BLOCKED_MEAN] + result[Const.ALLOWED_MEAN]
        result[Const.TREND_CHANGE_NORM_FACTOR] = result.apply(
            lambda d: 1
            if d[Const.TOTAL_MEAN] <= 0.0
            else (d[Const.TOTAL_BLOCKED] + d[Const.TOTAL_ALLOWED]) / d[Const.TOTAL_MEAN],
            axis=1,
        )

        # calc allowed ratio deviation from mean (with normalizing factor) & as percentage
        result[Const.TREND_CHANGE_DIFF] = (
            np.abs(result[Const.ALLOWED_RATIO_MEAN] - result[Const.ALLOW_RATIO])
            * result[Const.TREND_CHANGE_NORM_FACTOR]
        )

        result[Const.TREND_CHANGE_DIFF] = result[Const.TREND_CHANGE_DIFF] / result[Const.ALLOWED_RATIO_MEAN] * 100

        # calc score
        result[Const.TREND_CHANGE_SCORE] = result.apply(lambda d: np.minimum(d[Const.TREND_CHANGE_DIFF], 100), axis=1)

        DfLogger.log(
            result[
                [Const.TREND_CHANGE_SCORE, Const.TREND_CHANGE_DIFF, Const.TREND_CHANGE_NORM_FACTOR, Const.TOTAL_MEAN]
            ],
            site_id,
            direction,
            "stage2-score",
        )

        # return vm id & trend score only, reset index
        return result[[Const.TREND_CHANGE_SCORE]]

    def _execute_port_cardinality_stage(
        self, stage2_score: pd.DataFrame, site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        tag = f"{site_id}:{direction}"
        TrafficDrop.logger.info(f"STEP 3 ({tag}): Check new blocked ports")

        """
        V2: For all the vm result from stage 2:
        Count distinct port of blocked ingress flows (CardBDPort) into the vm list from stage 2 (V2)

        If CardBDPort(today+BL, V2) > CardBDPort(BL, V2) => flag
          (Basically if the trend of CardBDPort is ticking up => flag)

        B: Baseline Cardinality of DST Port, blocked flow (TODO: store the baseline)
        T: Today Cardinality of DST Port, blocked flow (baseL-today)
        (only src|dst blocked)

        For each vm
          sum(T) > sum(B)
          score (100)
          add score w/ weighted
        """

        vm_id = Direction.get_direction_vm_dimension(direction)

        # base & today interval
        baseline_interval = self._get_baseline_interval()
        baseline_to_now_interval = self._get_baseline_to_now_interval()

        # convert panda series of compute in to a list of string (for druid filter)
        vm_list = stage2_score.index.astype(str).values.tolist()

        baseline_cardinality = self._query.execute_port_cardinality_query(
            [baseline_interval], vm_list, site_id, direction
        )
        today_cardinality = self._query.execute_port_cardinality_query(
            [baseline_to_now_interval], vm_list, site_id, direction
        )

        # DfLogger.log(baseline_cardinality, site_id, direction, "baseline cardinality")
        # DfLogger.log(today_cardinality, site_id, direction, "today_cardinality")

        baseline_cardinality_sum_df = baseline_cardinality.groupby(vm_id).sum(numeric_only=True)
        today_cardinality_sum_df = today_cardinality.groupby(vm_id).sum(numeric_only=True)

        # DfLogger.log(baseline_cardinality_sum_df, site_id, direction, "baseline cardinality sum by vm")
        # DfLogger.log(today_cardinality_sum_df, site_id, direction, "today_cardinality sum by vm")

        # merge baseline sum (_B) with today sum (_T)
        result = pd.merge(
            baseline_cardinality_sum_df,
            today_cardinality_sum_df,
            on=vm_id,
            suffixes=(Const.BASELINE_SUFFIX, Const.TODAY_SUFFIX),
        )

        if len(result.index) == 0:
            TrafficDrop.logger.debug(
                "merge for today & baseline port cardinality is empty, returning empty df for stage3"
            )
            result[Const.CARDINALITY_SCORE] = None
            return result  # empty df

        result[Const.CARDINALITY_SCORE] = result.apply(
            lambda d: Const.NEW_PORT_MAX_SCORE
            if d[Const.TODAY_CARDINALITY_COUNT] > d[Const.BASELINE_CARDINALITY_COUNT]
            else 0.0,
            axis=1,
        )

        DfLogger.log(result, site_id, direction, "stage3-score")

        # return vm id & trend score only, reset index
        return result[[Const.CARDINALITY_SCORE]]

    def _combine_score(
        self,
        stage1_score: pd.DataFrame,
        stage2_score: pd.DataFrame,
        stage3_score: pd.DataFrame,
        user_config: UserConfigInfo,
        site: str,
        direction: TrafficDirection,
    ) -> pd.DataFrame:
        tag = f"{site}:{direction}"
        TrafficDrop.logger.info(f"Score ({tag}) : Calculating weighted score")

        vm_dimension = Direction.get_direction_vm_dimension(direction)

        # not all VM are available in every stages, so merge is done using outer join
        # and replacing NaN -> 0 (note: use explicit fillna to be safer)
        score1_2 = pd.merge(stage1_score, stage2_score, how="outer", on=vm_dimension)
        score1_2 = score1_2.fillna(0)
        combined_score = pd.merge(score1_2, stage3_score, how="outer", on=vm_dimension)
        combined_score = combined_score.fillna(0)

        if len(combined_score.index) == 0:
            TrafficDrop.logger.debug("combined score is empty")
            return combined_score  # empty df

        # calculate the weightage based on the constant weights (50/25/25)
        combined_score[Const.WEIGHTED_SCORE] = combined_score.apply(
            lambda d: d[Const.BLOCKED_STD_SCORE] * Const.SCORE_WEIGHTAGE[Const.BLOCKED_STD_SCORE] / 100
            + d[Const.TREND_CHANGE_SCORE] * Const.SCORE_WEIGHTAGE[Const.TREND_CHANGE_SCORE] / 100
            + d[Const.CARDINALITY_SCORE] * Const.SCORE_WEIGHTAGE[Const.CARDINALITY_SCORE] / 100,
            axis=1,
        )

        # post clean up of score
        TrafficDrop.logger.info(f"config settings: f{user_config.settings}")

        # clip the max to 100
        combined_score[Const.WEIGHTED_SCORE] = combined_score[Const.WEIGHTED_SCORE].clip(upper=Const.MAX_SCORE)

        # filter & determine silent
        sensitivity_setting = user_config.settings.get_sensitivity_meter_setting()
        combined_score = TrafficDrop._filter_low_score(combined_score, sensitivity_setting)
        combined_score = TrafficDrop._mark_as_silent(combined_score, sensitivity_setting)

        return combined_score

    @staticmethod
    def _filter_low_score(combined_score: pd.DataFrame, sensitivity_setting) -> pd.DataFrame:
        if Config.PERSIST_LOW_CONFIDENCE_EVENTS_AS_SILENT != "True":
            # only keep events with high enough score
            combined_score = combined_score[combined_score[Const.WEIGHTED_SCORE] >= sensitivity_setting.sensitivity]
            TrafficDrop.logger.info(
                f"after dropping low confidence events, num events to report is {len(combined_score.index)}"
            )

        return combined_score

    @staticmethod
    def _mark_as_silent(combined_score: pd.DataFrame, sensitivity_setting) -> pd.DataFrame:
        combined_score[Const.IS_SILENT] = combined_score[Const.WEIGHTED_SCORE] < sensitivity_setting.sensitivity

        return combined_score

    def _get_rule_id_for_compute(self, compute_id: str, site: str, direction: TrafficDirection) -> list[str]:
        all_rules = set()
        compute_rule_id_df = self._query.execute_rule_id_query(
            [self._get_day_start_to_now_interval()], compute_id, site, direction
        )

        """
        Rule Id DF (execute_rule_id_query)
        |    |   srcRuleIds | dstRuleIds   | timestamp                |
        |---:|-------------:|:-------------|:-------------------------|
        |  0 |         2002 | 2002         | 2023-05-23T00:00:00.000Z |
        |... |         ...  | ....         | 2023-05-23T01:00:00.000Z |
        """

        for _, row in compute_rule_id_df.iterrows():
            all_rules.add(row[Const.SRC_RULE_ID])
            all_rules.add(row[Const.DST_RULE_ID])

        return list(all_rules)

    def _get_policy_path(
        self, interval: Interval, site_id: str, direction: TrafficDirection, rule_ids: list[str]
    ) -> list[str]:
        TrafficDrop.logger.info(f"Finding policy paths for rule id {rule_ids}")
        df = self._query.execute_get_policy_path_for_rule_id(interval, site_id, direction, rule_ids)

        """
        rule, path DF (execute_get_policy_path_for_rule_id)
        |   rule_id | policy_path         | timestamp                |
        |----------:|:--------------------|:-------------------------|
        |      2002 | simulator/path/2002 | 2023-04-23T00:00:00.000Z |
        """
        return df[Const.POLICY_PATH].tolist()

    def _get_fw_change_event_log(
        self, interval: Interval, rule_ids: list[str], site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        # Replaces: Map fwChanges = fwRuleChangeReporter.getEventLog(interval, rule_ids, false);
        #
        # create 3 sets:
        #   oldest query (O): execute_fw_event_query (1st one, replaces executeQuery)
        #                     oldest_fw_within_baseline (interval: baseline_start-now)
        #   latest query (L): _get_policy_path (replaces getPolicyPathsForMpRules)
        #                     current_fw (interval: baseline_start-now)
        #   epoch query  (E): execute_fw_event_query (2nd one, replaces executeQuery)
        #                     fw_outside_of_baseline (interval: epoch-baseline_start)
        #
        # 1.  Query all the fw (and associated policy) from baseline_start till now (using rule id)
        #     note:
        #     if we only limit our query to interval: (baseline_start-now), some policy (e.g. P2)
        #     will not be in the 'oldest' set (we missed the oldest version of P2).
        #     A non-zerp set difference between 'latest' and 'oldest' implies the policy Pn lives
        #     between epoch-baseline_start. (see P2 below)
        #
        # time ----------|------------------------|------>>
        # epoch ---------| baseline_start --------| now
        #                  oldest query             latest query
        #                  P1 (O) ----------------| P1(L)         P1: Set(L) - S(O) = 0
        # P2 (oldest) ----------------------------| P2(L)         P2: Set(L) - S(O) != 0
        #                            P3 (O) ------| P3(L)
        # epoch query
        # (E) -----------|
        #
        # Pn = nth policy (e.g. Policy P1, Policy P2 etc.)
        #
        # 2a. The 1st execute_fw_event_query find all the oldest policy (but within baseline interval) for each rule id
        #     of interest (see step 1)
        #     The set difference between current_fw and fw_event_oldest_within_baseline_df will determine if we
        #     need to execute the 2nd execute_fw_event_query
        #
        # 2b. The 2nd execute_fw_event_query (ExecuteQuery) will pull out all those policy that was original created
        #     outside of the baseline interval (e.g. P2 was created in epoch-baseline_start)
        # 3.  add the result from 2a (oldest_fw_within_baseline) & 2b to complete the history of P1,P2,P3

        # 1. Get today Policy paths used by rule Ids
        """
        rule, path DF (execute_get_policy_path_for_rule_id)
        |   rule_id | policy_path         | timestamp                |
        |----------:|:--------------------|:-------------------------|
        |      2002 | simulator/path/2002 | 2023-04-23T00:00:00.000Z |
        """
        current_fw = self._get_policy_path(interval, site_id, direction, rule_ids)

        # 2a. Get oldest fw rule within baseline-now
        #   groupby(policypath, revision)[oldest_time, latest_displayname)
        """
        Firewall Event DF (execute_fw_event_query)
        |    | policy_path         |   display_name |    event_time |   revision | timestamp                |
        |---:|:--------------------|---------------:|--------------:|-----------:|:-------------------------|
        |  0 | simulator/path/2002 |           2002 | 1684369925374 |          0 | 2023-04-23T00:00:00.000Z |
        """
        fw_event_oldest_within_baseline_df = self._query.execute_fw_event_query(
            [interval], site_id, direction, current_fw
        )
        oldest_fw_within_baseline = fw_event_oldest_within_baseline_df[Const.POLICY_PATH]

        # difference between current_fw & fw_event_oldest_within_baseline
        #   return fw rule that was created "outside" of baseline-now (i.e. it was create between epoch-baseline_start)
        # TODO: add validation to make sure 'current_fw' set > 'oldest_fw_within_baseline'
        fw_outside_of_baseline = set(current_fw) - set(oldest_fw_within_baseline)

        # 2b: query rule create outside of the baseline (i.e. epoch-baseline_start)
        epoch_interval = self._get_epoch_to_baseline_start_interval()
        fw_event_epoch_to_baseline_start_df = self._query.execute_fw_event_query(
            [epoch_interval], site_id, direction, list(fw_outside_of_baseline)
        )

        # concat df & return
        result = pd.concat([fw_event_oldest_within_baseline_df, fw_event_epoch_to_baseline_start_df])

        DfLogger.log(result, site_id, direction, "full-fw-event")

        return result

    def _get_new_ports_for_event(self, compute_id: str, site: str, direction: TrafficDirection) -> set[str]:
        baseline_dstport = self._query.execute_get_dstport_query(
            [self._get_baseline_to_now_interval()], compute_id, site, direction
        )
        today_dstport = self._query.execute_get_dstport_query(
            [self._get_day_start_to_now_interval()], compute_id, site, direction
        )

        """
        DST Port DF (execute_get_dstport_query)
        |    |   dstPort | timestamp                |
        |---:|----------:|:-------------------------|
        |  0 |        56 | 2023-04-24T00:00:00.000Z |
        |  1 |        56 | 2023-04-25T00:00:00.000Z |
        |  2 |        54 | 2023-04-26T00:00:00.000Z |
        """
        baseline_dstport_set = set(baseline_dstport[Const.DST_PORT].tolist())
        today_dstport_set = set(today_dstport[Const.DST_PORT].tolist())

        return today_dstport_set - baseline_dstport_set

    def _create_event(
        self,
        compute_id: str,
        compute_row: pd.Series,
        site_id: str,
        direction: TrafficDirection,
        compute_to_ip_lookup: dict[str, str],
    ) -> EventInfo:
        # Mark detection event time (epoch mill)
        detection_time = int(self._now.timestamp() * 1000)

        # Execute TrafficDrop Report Query: replace executeVmDropCountReportQuery
        #   Find the allowedCount & blockedCount (baseline - now) & attach them as a data pt series
        allow_block_report_df = self._query.execute_base_traffic_drop_query(
            [self._get_baseline_to_now_interval()], site_id, direction, [compute_id]
        )

        DfLogger.log(allow_block_report_df, site_id, direction, "allow block report")

        # RuleChanges: replace getConfigChanges()
        #    Get rule changes in the "current interval" (for event) as compare to the baseline
        #      Execute DruidQuery for RuleId: select srcRuleIds, dstRuleIds from correlated_flow_viz where vm in "event"
        #    Find all src & dst Rule for the vm in event -> getEventLog
        #      PolicyChangeEvents: another query to find rulePath + revision
        #    note: instead of the java implementation of return a <timestamp, list<config_event>, a df of
        #          df[time, config_event] is easier since it all get converted to a API marker object
        #
        #    Set<String> mpRuleIds = getRuleIdsForEntity(vmId, CURRENT_INTERVAL, BASELINE_GRANULARITY, direction);
        #    Map fwChanges = fwRuleChangeReporter.getEventLog(interval, mpRuleIds, false);
        #  Druid doesn't offer a groupby-distinct-set aggregator
        today_compute_rules = self._get_rule_id_for_compute(compute_id, site_id, direction)
        fw_change_event_df = self._get_fw_change_event_log(
            self._get_baseline_to_now_interval(), today_compute_rules, site_id, direction
        )

        # getNewPorts
        #    Goal:
        #    B = getDstPortSet(baseline_interval)
        #    T = getDstPortSet(today)
        #    D = B-T
        #    Use D in event
        new_ports = self._get_new_ports_for_event(compute_id, site_id, direction)

        # getOppositeEndVms map
        #    Goal: for all events in the input, find the opposite side VM
        #    DruidQuery: ...
        opposite_compute = self._query.execute_opposite_compute_query(
            self._get_baseline_to_now_interval(), compute_id, site_id, direction
        )

        # Generate event: replace TrafficDropAnomalyEventBuilder
        event_info = (
            Builder()
            .detector_type(self.detector_name())
            .compute_id(compute_id)
            .description(direction, compute_id)
            .direction(direction)
            .detection_time(detection_time)
            .silent(compute_row[Const.IS_SILENT])
            .org_id(Config.ORG_ID)
            .site_id(site_id)
            .allowed_counts(allow_block_report_df)
            .blocked_counts(allow_block_report_df)
            .config_changes(fw_change_event_df)
            .compute_ip_map(compute_to_ip_lookup)
            .new_ports(list(new_ports))
            .src_computes(compute_id, opposite_compute[Const.OPPOSITE_SIDE].drop_duplicates().tolist())
            .dst_computes(compute_id, opposite_compute[Const.OPPOSITE_SIDE].drop_duplicates().tolist())
            .hot_spot([compute_id])
            .viz_filter()
            .confidence(compute_row[Const.WEIGHTED_SCORE])
            .build()
        )

        return event_info

    def _get_detector_run_status(
        self, status_label: DetectorStatusLabel, site_id: str, last_successful_run_time: int, message: str = ""
    ) -> DetectorRunStatus:
        TrafficDrop.logger.debug(f"Setting detector run status for {self.detector_name().name}")
        max_baseline_duration_days = self._baseline_service.get_max_baseline_duration_days(
            self.detector_name().name, Config.ORG_ID, site_id
        )
        remaining_baseline_duration_days = max(int(Config.MIN_BASELINE_DURATION_DAYS) - max_baseline_duration_days, 0)
        status_label = (
            DetectorStatusLabel.NOT_ENOUGH_BASELINE
            if (status_label == DetectorStatusLabel.SUCCESS and remaining_baseline_duration_days > 0)
            else status_label
        )

        return DetectorRunStatus(
            status_label=status_label,
            last_successful_run_time=last_successful_run_time,
            max_baseline_duration_days=max_baseline_duration_days,
            remaining_baseline_duration_days=remaining_baseline_duration_days,
            message=message,
        )
