import json
from typing import Self

import numpy as np
import pandas as pd

import openapi.models.base_anomaly_event_detail  # needed to avoid circular import for any subclasses  # noqa: F401
from common.utils.common_model import DetectorType
from common.utils.event.data import (
    Compute,
    ComputeType,
    DataFlow,
    EventInfo,
    EventKey,
    EventMetadata,
    IpMask,
    JsonValue,
    ListString,
    MergePlan,
    MergePolicy,
    Score,
    Tactic,
    Technique,
)
from common.utils.vizfilter.core import (
    CompositeVizFilter,
    Condition,
    Criteria,
    VizFilter,
    VizFilterOption,
)
from openapi.models.series import Series
from openapi.models.traffic_drop_event_detail import TrafficDropEventDetail
from openapi.models.traffic_drop_event_detail_marker import TrafficDropEventDetailMarker
from trafficdrop.detector.common_types import TrafficDirection
from trafficdrop.detector.config import Config
from trafficdrop.detector.const import Const
from vmware.nsx.intelligence.nta.v1 import nta_pb2


class Builder:
    """
    convert varies metrics about traffic drop event into the API object
    """

    # TODO: consider moving these outside this module

    # These are enum values used in TrafficDropEventDetails properties
    # The openapi doesn't generate an explicit Enum value but only generates a validator for it
    INCOMING = "INCOMING"
    OUTGOING = "OUTGOING"
    RULE_CHANGE = "RULE_CHANGE"
    GROUP_CHANGE = "GROUP_CHANGE"

    _compute_id: str
    _description: str
    _detector_type: DetectorType
    _detection_time: int
    _direction: str  # API Object: INCOMING vs OUTGOING
    _src_computes: list[str]
    _dst_computes: list[str]
    _org_id: str
    _site_id: str
    _silent: bool
    _hotspots: list[str]
    _new_ports: list[str]
    _markers: list[TrafficDropEventDetailMarker]
    _confidence: int
    _final_viz_filter: CompositeVizFilter
    _compute_to_ip: dict[str, str]
    _allowed_count_series: Series
    _blocked_count_series: Series

    def compute_id(self, compute_id: str) -> Self:
        self._compute_id = compute_id
        return self

    def description(self, direction: TrafficDirection, compute_id: str) -> Self:
        self._description = (
            f"Anomalous behavior in outgoing traffic from VM {compute_id}"
            if direction is TrafficDirection.SRC
            else f"Anomalous behavior in incoming traffic to VM {compute_id}"
        )
        return self

    def detector_type(self, detector_type: DetectorType) -> Self:
        self._detector_type = detector_type
        return self

    def org_id(self, org_id: str) -> Self:
        self._org_id = org_id
        return self

    def site_id(self, site_id: str) -> Self:
        self._site_id = site_id
        return self

    def config_changes(self, fw_change_event_df: pd.DataFrame) -> Self:
        self._markers = []
        for _, row in fw_change_event_df.iterrows():
            self._markers.append(
                TrafficDropEventDetailMarker(
                    timestamp=row[Const.EVENT_TIME],
                    config_entity_id=row[Const.POLICY_PATH],
                    config_entity_display_name=row[Const.DISPLAY_NAME],
                    config_change_type=Builder.RULE_CHANGE,
                )
            )
        return self

    def blocked_counts(self, compute_block_allow_df: pd.DataFrame) -> Self:
        # porting note: we use the mean from the input df (but not mean from input as in the older code)
        ms_denominator = 10**6  # timestamp should be in ms
        self._blocked_count_series = Series(
            data=compute_block_allow_df[Const.TOTAL_BLOCKED].astype("float").to_list(),
            labels=(pd.to_datetime(compute_block_allow_df[Const.TIMESTAMP]).astype("int") // ms_denominator)
            .astype(str)
            .to_list(),
            mean=compute_block_allow_df[Const.TOTAL_BLOCKED].mean(),
        )
        return self

    def allowed_counts(self, compute_block_allow_df: pd.DataFrame) -> Self:
        # porting note: we use the mean from the input df (but not mean from input as in the older code)
        ms_denominator = 10**6  # timestamp should be in ms
        self._allowed_count_series = Series(
            data=compute_block_allow_df[Const.TOTAL_ALLOWED].astype("float").to_list(),
            labels=(pd.to_datetime(compute_block_allow_df[Const.TIMESTAMP]).astype("int") // ms_denominator)
            .astype(str)
            .to_list(),
            mean=compute_block_allow_df[Const.TOTAL_ALLOWED].mean(),
        )
        return self

    def compute_ip_map(self, compute_to_ip: dict[str, str]) -> Self:
        self._compute_to_ip = compute_to_ip
        return self

    def confidence(self, score) -> Self:
        self._confidence = np.floor(score).astype(int)  # score need to be int
        return self

    def new_ports(self, new_ports: list[str]) -> Self:
        self._new_ports = new_ports
        return self

    def src_computes(self, compute_id: str, opposite_compute: list[str]) -> Self:
        self._src_computes = opposite_compute if self._direction is TrafficDirection.DST else [compute_id]
        return self

    def dst_computes(self, compute_id: str, opposite_compute: list[str]) -> Self:
        self._dst_computes = [compute_id] if self._direction is TrafficDirection.DST else opposite_compute
        return self

    def hot_spot(self, hotspots: list[str]) -> Self:
        # same as a single compute_id
        self._hotspots = hotspots
        return self

    def direction(self, direction: TrafficDirection) -> Self:
        self._direction = Builder.INCOMING if direction == TrafficDirection.DST else Builder.OUTGOING
        return self

    def detection_time(self, detection_time: int) -> Self:
        self._detection_time = detection_time
        return self

    def silent(self, is_silent: bool) -> Self:
        self._silent = is_silent
        return self

    def viz_filter(self) -> Self:
        # filter hotspot OR
        vm_viz_filter = VizFilter(
            key=VizFilterOption.VM_ID, values=self._hotspots, criteria=Criteria.EQ, condition=Condition.OR
        )

        flow_filter = VizFilter(
            key=VizFilterOption.FLOW_TYPE, values=[VizFilter.BLOCKED], criteria=Criteria.EQ, condition=Condition.OR
        )

        self._final_viz_filter = CompositeVizFilter([vm_viz_filter, flow_filter], Condition.AND)
        return self

    def build_details(self) -> TrafficDropEventDetail:
        return TrafficDropEventDetail(
            marker=self._markers,
            drop_count=self._blocked_count_series,
            allow_count=self._allowed_count_series,
            new_ports=self._new_ports,
            direction=self._direction,
            # TODO: BaseAnomalyEventDetail report_type should be identical to the detector name
            #       But it is name 'TRAFFIC_DROP' not 'NETWORK_TRAFFIC_DROP'
            report_type=Const.REPORT_TYPE,
        )

    def build(self) -> EventInfo:
        # limit src and dst computes for scale reasons, too large of metadata will cause kafka to reject the message
        self._src_computes = self._src_computes[: int(Config.EVENT_METADATA_COMPUTE_LIMIT)]
        self._dst_computes = self._dst_computes[: int(Config.EVENT_METADATA_COMPUTE_LIMIT)]
        self._hotspots = self._hotspots[: int(Config.EVENT_METADATA_COMPUTE_LIMIT)]

        # metadata
        event_metadata = EventMetadata(
            kv_store={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): ListString(self._src_computes),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): ListString(self._dst_computes),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_HOTSPOTS): ListString(self._hotspots),
                EventKey.EVENT_METADATA_REPORT_KEY: JsonValue(json.dumps(self.build_details().to_dict())),
                EventKey.EVENT_METADATA_DESCRIPTION_KEY: JsonValue(self._description),
                EventKey.EVENT_METADATA_VIZ_FILTER_KEY: JsonValue(self._final_viz_filter.to_json_str()),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): ListString(
                    [self._compute_to_ip[compute] for compute in self._compute_to_ip if compute in self._src_computes]
                ),
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): ListString(
                    [self._compute_to_ip[compute] for compute in self._compute_to_ip if compute in self._dst_computes]
                ),
            }
        )

        # merge_plan
        merge_plan = MergePlan(
            # porting note: the old code is iterating through the dst/src compute as key, but we now use the
            #               compute_id for each builder invocation
            dedup_key="{}_{}_{}_{}".format(
                Const.NSX_I, str(DetectorType.NETWORK_TRAFFIC_DROP), str(self._direction), self._compute_id
            ),
            metadata_key_merge_operation={
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_COMPUTES): MergePolicy.SET_UNION_WITH_MAX
                if self._direction is Builder.INCOMING
                else MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_COMPUTES): MergePolicy.KEEP_ORIGINAL
                if self._direction is Builder.INCOMING
                else MergePolicy.SET_UNION_WITH_MAX,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_SRC_IPS): MergePolicy.KEEP_ORIGINAL,
                nta_pb2.MetadataKey.Name(nta_pb2.MetadataKey.METADATA_KEY_DST_IPS): MergePolicy.KEEP_ORIGINAL,
            },
        )

        affected_compute_ip = self._compute_to_ip[self._compute_id] if self._compute_id in self._compute_to_ip else None
        affected_compute = Compute(
            compute_type=ComputeType.COMPUTE_TYPE_VM,
            ip_address=affected_compute_ip,
            compute_id=self._compute_id,
            fqdn=None,
            site_id=self._site_id,
            ip_mask=IpMask.PRIVATE,
        )

        if self._direction == self.OUTGOING:
            # affected compute is the client/attacker
            is_client = is_attacker = True
            # other computes are all the dst computes
            other_computes = self._dst_computes
        else:
            # affected compute is the target/server
            is_client = is_attacker = False
            # other computes are all the src computes
            other_computes = self._src_computes

        data_flows = []
        for other_compute_id in other_computes:
            other_compute_ip = self._compute_to_ip.get(other_compute_id)
            other_compute = Compute(
                compute_type=ComputeType.COMPUTE_TYPE_VM,
                ip_address=other_compute_ip,
                compute_id=other_compute_id,
                fqdn=None,
                site_id=self._site_id,
                ip_mask=IpMask.PRIVATE,
            )
            data_flows.append(
                DataFlow(
                    src_compute=(affected_compute if self._direction == self.OUTGOING else other_compute),
                    dst_compute=(other_compute if self._direction == self.OUTGOING else affected_compute),
                    # We don't have any transport protocol information at this stage
                    transport_protocols=None,
                    application_protocols=None,
                    # We don't have port information at this stage
                    server_ports=None,
                    # We don't have flows' timestamp information at this stage: use the detection
                    # time
                    start_time=self._detection_time,
                    end_time=self._detection_time,
                )
            )

        return EventInfo(
            version="v1.1",
            detector_name=self._detector_type,
            detection_start_time_epoch_milli=self._detection_time,
            detection_end_time_epoch_milli=self._detection_time,
            score=Score(self._confidence, severity=20),
            event_metadata=event_metadata,
            merge_plan=merge_plan,
            mitre_tactic=Tactic.OTHER,
            mitre_technique=Technique.OTHER,
            silent=self._silent,
            site_ids=[self._site_id],
            org_id=self._org_id,
            affected_compute=affected_compute,
            is_attacker=is_attacker,
            is_client=is_client,
            data_flows=data_flows,
        )
