import uuid

import pandas as pd
from common.utils import logging
from common.utils.common_model import Interval
from common.utils.flow.core import FlowService
from common.utils.flow.data import FlowGroupByQuery, FlowVirtualColumn
from pydruid.utils.aggregators import cardinality, longsum
from pydruid.utils.filters import Bound, Dimension, Filter
from pydruid.utils.postaggregator import Field
from trafficdrop.detector.common_types import TrafficDirection
from trafficdrop.detector.config import Config
from trafficdrop.detector.const import Const
from trafficdrop.detector.utils import DfLogger, Direction


class DruidQuery:
    logger = logging.get_common_logger(__name__)

    def __init__(self, detector_name: str, flow_service: FlowService):
        self._flow_service = flow_service
        self._detector_name = detector_name

    def execute_base_traffic_drop_query(
        self,
        intervals: list[Interval],
        site_id: str,
        direction: TrafficDirection,
        included_vms: list[str] = [],
        excluded_vms: list[str] = [],
    ) -> pd.DataFrame:
        """
        execute the druid query to look for block count (+other related metrics)

        :param intervals:
        :param site_id:
        :param direction:
        :param included_vms:
        :param excluded_vms:
        :return:
        """
        dimension_name = Direction.get_direction_vm_dimension(direction)
        filter_compute_type = Const.SRC_COMPUTE_TYPE if direction == TrafficDirection.SRC else Const.DST_COMPUTE_TYPE
        filter_site = Const.SRC_SITE_ID if direction == TrafficDirection.SRC else Const.DST_SITE_ID
        filter_info_present = (
            Const.IS_SRC_INFO_PRESENT if direction == TrafficDirection.SRC else Const.IS_DST_INFO_PRESENT
        )

        filters = [
            Dimension(filter_compute_type) == Const.VM_COMPUTE_TYPE,
            Dimension(filter_site) == site_id,
            Dimension(filter_info_present) == True,  # noqa
        ]
        if included_vms != []:
            filters.append(Filter(type="in", dimension=dimension_name, values=included_vms))
        if excluded_vms != []:
            filters.append(~Filter(type="in", dimension=dimension_name, values=excluded_vms))

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=intervals,
            dimensions=[dimension_name],
            virtual_columns=[
                FlowVirtualColumn(Const.EXT_SRC_ALLOWED, Const.EXT_SRC_ALLOWED_VCOL),
                FlowVirtualColumn(Const.EXT_DST_ALLOWED, Const.EXT_DST_ALLOWED_VCOL),
            ],
            aggregations={
                Const.SRC_FW_BLOCKED: longsum(Const.SRC_FW_BLOCKED),
                Const.DST_FW_BLOCKED: longsum(Const.DST_FW_BLOCKED),
                Const.SRC_FW_ALLOWED: longsum(Const.SRC_FW_ALLOWED),
                Const.DST_FW_ALLOWED: longsum(Const.DST_FW_ALLOWED),
                Const.EXT_SRC_ALLOWED: longsum(Const.EXT_SRC_ALLOWED),
                Const.EXT_DST_ALLOWED: longsum(Const.EXT_DST_ALLOWED),
            },
            post_aggregations={
                Const.COMBINED_SRC_ALLOW: Field(Const.SRC_FW_ALLOWED) + Field(Const.EXT_SRC_ALLOWED),
                Const.COMBINED_DST_ALLOW: Field(Const.DST_FW_ALLOWED) + Field(Const.EXT_DST_ALLOWED),
                Const.TOTAL_BLOCKED: Field(Const.SRC_FW_BLOCKED) + Field(Const.DST_FW_BLOCKED),
                Const.TOTAL_ALLOWED: Field(Const.COMBINED_SRC_ALLOW) + Field(Const.COMBINED_DST_ALLOW),
                Const.TOTAL: Field(Const.TOTAL_BLOCKED) + Field(Const.TOTAL_ALLOWED),
                Const.BLOCK_RATIO: Field(Const.TOTAL_BLOCKED) / Field(Const.TOTAL),
                Const.ALLOW_RATIO: Field(Const.TOTAL_ALLOWED) / Field(Const.TOTAL),
            },
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:{direction}:blocked:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing baseline query for {direction}")
        DruidQuery.logger.debug(f"{query.pprint()}")
        result = self._flow_service.group_by(query)

        # handle empty dataframe (druid group_by can't return a empty df with columns defined)
        if len(result.columns) == 0:
            empty_df = pd.DataFrame(
                columns=[
                    dimension_name,
                    Const.DST_FW_ALLOWED,
                    Const.EXT_DST_ALLOWED,
                    Const.COMBINED_DST_ALLOW,
                    Const.ALLOW_RATIO,
                    Const.TOTAL_BLOCKED,
                    Const.SRC_FW_ALLOWED,
                    Const.TOTAL_ALLOWED,
                    Const.SRC_FW_BLOCKED,
                    Const.COMBINED_SRC_ALLOW,
                    Const.DST_FW_BLOCKED,
                    Const.BLOCK_RATIO,
                    Const.TOTAL,
                    Const.EXT_SRC_ALLOWED,
                    Const.TIMESTAMP,
                ]
            )
            result = empty_df

        result.set_index(dimension_name, inplace=True)

        DfLogger.log(result, site_id, direction, f"blocked:{intervals}")

        return result

    def execute_port_cardinality_query(
        self, intervals: list[Interval], vm_list: list[str], site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        vm_dimension = Direction.get_direction_vm_dimension(direction)

        # bound(srcFwBlocked) > 0 or bound(dstFwdBlocked) > 0
        fw_filters = [
            Bound(Const.SRC_FW_BLOCKED, "0", lowerStrict=True),
            Bound(Const.DST_FW_BLOCKED, "0", lowerStrict=True),
        ]
        filters = [Filter(type="or", fields=fw_filters), Filter(type="in", dimension=vm_dimension, values=vm_list)]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_ALL_GRANULARITY,
            intervals=intervals,
            dimensions=[vm_dimension],
            virtual_columns=[],
            aggregations={Const.CARDINALITY_COUNT: cardinality(Const.DST_PORT)},
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:dstport-cardinality:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing {direction} dstport cardinality query")
        DruidQuery.logger.debug(f"{query.pprint()}")
        result = self._flow_service.group_by(query)

        if len(result.columns) == 0:
            result = pd.DataFrame(columns=[vm_dimension, Const.CARDINALITY_COUNT])
        result.set_index(vm_dimension, inplace=True)

        DfLogger.log(result, site_id, direction, "dst port cardinality")

        return result

    def execute_rule_id_query(
        self, intervals: list[Interval], vm: str, site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        vm_dimension = Direction.get_direction_vm_dimension(direction)

        # bound(srcFwBlocked) > 0 or bound(dstFwdBlocked) > 0
        fw_filters = [
            Bound(Const.SRC_FW_BLOCKED, "0", lowerStrict=True),
            Bound(Const.DST_FW_BLOCKED, "0", lowerStrict=True),
        ]
        filters = [Filter(type="or", fields=fw_filters), Filter(type="in", dimension=vm_dimension, values=[vm])]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=intervals,
            dimensions=[Const.SRC_RULE_ID, Const.DST_RULE_ID],
            virtual_columns=[],
            aggregations={},
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:compute-ruleId:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing {direction} compute:{vm} rule id query")
        DruidQuery.logger.debug(f"{query.pprint()}")
        result = self._flow_service.group_by(query)

        if len(result.columns) == 0:
            result = pd.DataFrame(columns=[Const.SRC_RULE_ID, Const.DST_RULE_ID])

        DfLogger.log(result, site_id, direction, "rule-id")

        return result

    def execute_get_policy_path_for_rule_id(
        self, interval: Interval, site_id: str, direction: TrafficDirection, rule_ids: list[str]
    ) -> pd.DataFrame:
        filters = [
            Dimension(Const.CONFIG_TYPE) == Const.MANAGER_DFW_RULE,
            Filter(type="in", dimension=Const.RULE_ID, values=rule_ids),
        ]

        query = FlowGroupByQuery(
            datasource=Config.CONFIG_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_ALL_GRANULARITY,
            intervals=[interval],
            dimensions=[Const.RULE_ID],
            virtual_columns=[],
            aggregations={
                # replace stringlast w/ stringLastNsx:
                #   Const.POLICY_PATH: stringlast(Const.POLICY_PATH_FROM_TAG)
                Const.POLICY_PATH: {"type": Const.STRING_LAST_NSX, "fieldName": Const.POLICY_PATH_FROM_TAG}
            },
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:policypath-ruleId:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing policy path for rule id: {rule_ids}")
        result = self._flow_service.group_by(query)

        if len(result.columns) == 0:
            result = pd.DataFrame(columns=[Const.RULE_ID, Const.POLICY_PATH, Const.TIMESTAMP])
        result.set_index("rule_id", inplace=True)

        DfLogger.log(result, site_id, direction, "policypath-ruleid")

        return result

    def execute_fw_event_query(
        self, intervals: list[Interval], site_id: str, direction: TrafficDirection, included_paths: list[str] = []
    ) -> pd.DataFrame:
        filters = [
            Dimension(Const.CONFIG_TYPE) == Const.POLICY_DFW_RULE,
            Filter(type="in", dimension=Const.POLICY_PATH, values=included_paths),
        ]

        query = FlowGroupByQuery(
            datasource=Config.INTENT_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_ALL_GRANULARITY,
            intervals=intervals,
            dimensions=[Const.POLICY_PATH, Const.REVISION],
            virtual_columns=[],
            aggregations={
                # longFirst is missing in the library
                Const.EVENT_TIME: {"type": "longFirst", "fieldName": Const.TIME},
                # replace stringlast w/ stringLastNsx:
                #   Const.DISPLAY_NAME: stringlast(Const.DISPLAY_NAME)
                Const.DISPLAY_NAME: {"type": Const.STRING_LAST_NSX, "fieldName": Const.DISPLAY_NAME},
            },
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:fw-event-log:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing fw event log: {included_paths}")
        result = self._flow_service.group_by(query)

        if len(result.columns) == 0:
            result = pd.DataFrame(
                columns=[Const.EVENT_TIME, Const.DISPLAY_NAME, Const.POLICY_PATH, Const.REVISION, Const.TIMESTAMP]
            )
        result.set_index(Const.EVENT_TIME)

        DfLogger.log(result, site_id, direction, "fw-event-log")

        return result

    def execute_get_dstport_query(
        self, intervals: list[Interval], compute_id: str, site_id: str, direction: TrafficDirection
    ):
        vm_dimension = Direction.get_direction_vm_dimension(direction)

        # bound(srcFwBlocked) > 0 or bound(dstFwdBlocked) > 0
        fw_filters = [
            Bound(Const.SRC_FW_BLOCKED, "0", lowerStrict=True),
            Bound(Const.DST_FW_BLOCKED, "0", lowerStrict=True),
        ]
        filters = [Filter(type="or", fields=fw_filters), Filter(type="in", dimension=vm_dimension, values=[compute_id])]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=intervals,
            dimensions=[Const.DST_PORT],
            virtual_columns=[],
            aggregations={},
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:dstport-set:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing dstport-set: {compute_id}")
        result = self._flow_service.group_by(query)

        if len(result.columns) == 0:
            result = pd.DataFrame(columns=[Const.DST_PORT, Const.TIMESTAMP])

        DfLogger.log(result, site_id, direction, "dstport-set")

        return result

    def execute_opposite_compute_query(
        self, interval: Interval, compute_id: str, site_id: str, direction: TrafficDirection
    ) -> pd.DataFrame:
        DruidQuery.logger.info("Get opposite & VM")

        vm_dimension = Direction.get_direction_vm_dimension(direction)
        opposite_compute_dimension = Direction.get_opposite_compute_dimension(direction)

        # bound(srcFwBlocked) > 0 or bound(dstFwdBlocked) > 0
        fw_filters = [
            Bound(Const.SRC_FW_BLOCKED, "0", lowerStrict=True),
            Bound(Const.DST_FW_BLOCKED, "0", lowerStrict=True),
        ]
        filters = [Filter(type="or", fields=fw_filters), Filter(type="in", dimension=vm_dimension, values=[compute_id])]

        query = FlowGroupByQuery(
            datasource=Config.FLOW_DATA_SOURCE,
            granularity=Config.FLOW_QUERY_GRANULARITY,
            intervals=[interval],
            dimensions=[opposite_compute_dimension],
            virtual_columns=[],
            aggregations={
                # replace stringlast w/ stringLastNsx:
                #   Const.OPPOSITE_SIDE: stringlast(opposite_compute_dimension)
                Const.OPPOSITE_SIDE: {"type": Const.STRING_LAST_NSX, "fieldName": opposite_compute_dimension}
            },
            post_aggregations={},
            filter=Filter(type="and", fields=filters),
            context={"queryId": f"NTA:{self._detector_name}:opposite-vm:{uuid.uuid4()}"},
        )

        DruidQuery.logger.info(f"executing query opposite-vm: {compute_id}")
        result = self._flow_service.group_by(query)

        if len(result.columns) == 0:
            result = pd.DataFrame(columns=[opposite_compute_dimension, Const.OPPOSITE_SIDE, Const.TIMESTAMP])

        # filter out "External" & "Internal"
        result = result.loc[
            (result[Const.OPPOSITE_SIDE] != Const.EXTERNAL_VM_ID)
            & (result[Const.OPPOSITE_SIDE] != Const.INTERNAL_VM_ID)
        ]

        DfLogger.log(result, site_id, direction, "opposite-vm")

        return result
