from dataclasses import dataclass
from datetime import datetime
from enum import Enum, auto

import pandas as pd
from common.utils import logging
from dateutil import parser as dateparser
from trafficdrop.detector.common_types import TrafficDirection
from trafficdrop.detector.const import Const
from trafficdrop.detector.config import Config


class TrafficDropConfigEventType(Enum):
    """
    replaces ConfigEventType in original java implementation
    This represent the type of configuration change associated with the drop traffic
    """
    POLICY_FW_RULE_EVENT = auto()
    VM_GROUP_EVENT = auto()


@dataclass
class TrafficDropConfigEvent:
    """
    replaces BaseConfigEvent: used for creating markers for the event chart
    """
    policy_path: str
    display_name: str
    type: TrafficDropConfigEventType
    timestamp: int


class DfLogger:
    logger = logging.get_common_logger(__name__)
    unformatted_logger = logging.get_unformatted_logger(__name__)

    @staticmethod
    def log(df: pd.DataFrame, site_id: str, direction: TrafficDirection, message: str = ""):

        number_of_log_lines = int(Config.NUMBER_OF_LINES_OF_LOGS)
        with pd.option_context('display.max_rows', None,
                               'display.max_columns', None,
                               'display.precision', 3):
            result_df = df.iloc[:number_of_log_lines].to_markdown() if not df.empty else []
            DfLogger.unformatted_logger.debug("=====================================================================")
            DfLogger.logger.debug(f"{message}-{direction.name}-{site_id} result:")
            DfLogger.unformatted_logger.debug(f"{result_df}")
            DfLogger.unformatted_logger.debug("=====================================================================")


class Direction:
    @staticmethod
    def get_direction_vm_dimension(direction: TrafficDirection) -> str:
        return Const.SRC_VM_ID if direction == TrafficDirection.SRC else Const.DST_VM_ID

    @staticmethod
    def get_opposite_direction(direction: TrafficDirection) -> TrafficDirection:
        return TrafficDirection.SRC if direction == TrafficDirection.DST else TrafficDirection.DST

    @staticmethod
    def get_opposite_compute_dimension(direction: TrafficDirection) -> str:
        opposite_direction = Direction.get_opposite_direction(direction)
        return Direction.get_direction_vm_dimension(opposite_direction)


class IsoDateTimeParseError(Exception):
    pass


class Time:
    @staticmethod
    def parse_iso_datetime(isoformat_datetime: str) -> datetime:
        try:
            return dateparser.parse(isoformat_datetime)
        except dateparser.ParserError as e:
            raise IsoDateTimeParseError(
                f"unable to parse timestamp str{isoformat_datetime}") from e
