import io
import unittest
from datetime import date, datetime, timedelta, timezone
from pickle import dump
from unittest.mock import Mock

import pandas as pd

from common.utils.baseline.core import BaselineService
from common.utils.baseline.data import PartialBaseline
from common.utils.checkpoint.core import CheckpointService
from common.utils.common_model import Interval
from common.utils.detector.core import DetectorType
from common.utils.event.core import EventService
from common.utils.event.data import (
    Compute,
    ComputeType,
    DataFlow,
    IpMask,
    Score,
)
from common.utils.flow.core import FlowService
from common.utils.lookup.core import LookupService
from common.utils.userconfig.core import UserConfigService
from common.utils.userconfig.data import (
    AnomalyExclusionSetting,
    AnomalyExclusionSettings,
    DetectorRunStatus,
    DetectorSettings,
    DetectorStatusLabel,
    ExcludedItemsInfo,
    ExclusionLabel,
    ExclusionTypeInfo,
    SensitivityMeterSetting,
    UserConfigInfo,
)
from trafficdrop.detector.baseline import Baseline
from trafficdrop.detector.config import Config
from trafficdrop.detector.core import TrafficDirection, TrafficDrop, TrafficDropBaseline
from trafficdrop.detector.event import Builder


class TestCore(unittest.TestCase):
    def setUp(self):
        self._mock_flow_service = Mock(FlowService)
        self._mock_lookup_service = Mock(LookupService)
        self.detector = TrafficDrop(
            Mock(BaselineService),
            Mock(CheckpointService),
            Mock(UserConfigService),
            Mock(EventService),
            self._mock_flow_service,
            self._mock_lookup_service,
            datetime.now(),
        )

    def test_detector_name(self):
        self.assertEqual(DetectorType.NETWORK_TRAFFIC_DROP, self.detector.detector_name())

    @staticmethod
    def setupSampleDf(baseline, index_name: str):
        df = pd.DataFrame(baseline, columns=[index_name, "cardinalityCount"])
        df.set_index(index_name, inplace=True)

        return df

    @staticmethod
    def createTrafficDropBaseline(src_df: pd.DataFrame, dst_df: pd.DataFrame):
        src_buffer = io.BytesIO()
        src_df.to_parquet(src_buffer)
        src_buffer.seek(0)  # rewind the buffer

        dst_buffer = io.BytesIO()
        dst_df.to_parquet(dst_buffer)
        dst_buffer.seek(0)  # rewind the buffer

        return TrafficDropBaseline(src_buffer, dst_buffer)

    def test_combine_baseline(self):
        # src baseline for 3 days for 3 VM
        src_baselines = [
            [
                ["vmA", 1, 0, 11, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vmB", 2, 0, 22, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vmC", 3, 0, 33, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ],
            [
                ["vmA", 1, 0, 11, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vmB", 2, 0, 22, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vmC", 3, 0, 33, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
            ],
            [
                ["vmA", 1, 0, 11, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vmB", 2, 0, 22, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vmC", 3, 0, 33, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
            ],
        ]

        src_baselines = [
            pd.DataFrame(b, columns=["srcVmId", "srcFwBlocked", "dstFwBlocked", "Random", "timestamp"])
            for b in src_baselines
        ]

        dst_baselines = [
            [
                ["vmD", 0, 4, 44, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vmE", 0, 5, 55, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
                ["vmF", 0, 6, 66, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ],
            [
                ["vmD", 0, 5, 44, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vmE", 0, 6, 55, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
                ["vmF", 0, 7, 66, datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
            ],
            [
                ["vmD", 0, 4, 44, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vmE", 0, 7, 55, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
                ["vmF", 0, 9, 66, datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
            ],
        ]

        dst_baselines = [
            pd.DataFrame(b, columns=["dstVmId", "srcFwBlocked", "dstFwBlocked", "Random", "timestamp"])
            for b in dst_baselines
        ]

        print(src_baselines)
        print(dst_baselines)

        partial_baselines = self.create_partial_baseline(src_baselines, dst_baselines)
        print(partial_baselines)

        combined_baseline_byteio = self.detector._combine_baseline_objects(partial_baselines)
        if combined_baseline_byteio is not None:
            combined_baseline_byteio.seek(0)

            src_baseline, dst_baseline = Baseline.get_combined_df_from(combined_baseline_byteio)

            # get the original input 3 x 3 rows back
            self.assertEqual(src_baseline.shape[0], 9)
            # get the original input 3 x 3 rows back
            self.assertEqual(dst_baseline.shape[0], 9)
        else:
            self.fail("baseline buffer should not be None")

    def create_partial_baseline(
        self, src_partial_baseline_df: list[pd.DataFrame], dst_partial_baseline_df: list[pd.DataFrame]
    ) -> list[PartialBaseline]:
        # create a list of partial baseline
        partial_baselines = []

        # note: the src df & dst df are of same size
        for p in list(zip(src_partial_baseline_df, dst_partial_baseline_df)):
            sbuf = io.BytesIO()
            dbuf = io.BytesIO()

            src_pbaseline = p[0]
            dst_pbaseline = p[1]
            src_pbaseline["timestamp"] = pd.to_datetime(src_pbaseline["timestamp"], format="%Y-%m-%d %H:%M:%S")
            dst_pbaseline["timestamp"] = pd.to_datetime(dst_pbaseline["timestamp"], format="%Y-%m-%d %H:%M:%S")

            src_pbaseline.to_parquet(sbuf)
            dst_pbaseline.to_parquet(dbuf)

            sbuf.seek(0)
            dbuf.seek(0)

            # interval (all the entries in the same partial baseline has the same start date
            # pick anyone
            start_datetime = src_pbaseline["timestamp"][0].to_pydatetime()
            interval = Interval(start_datetime, start_datetime + timedelta(days=1))

            # combine query results
            buffer = io.BytesIO()
            pbaseline = TrafficDropBaseline(sbuf, dbuf)
            dump(pbaseline, buffer)
            partial_baselines.append(PartialBaseline(interval, buffer))
            buffer.seek(0)

        return partial_baselines

    def _getDefaultUserConfig(self) -> UserConfigInfo:
        return UserConfigInfo(
            site_id="site1",
            name="TDROP",
            org_id="org_1",
            version="1",
            last_modified_time=1234,
            enabled=True,
            silent=False,
            source="source",
            block=False,
            block_confidence_threshold=1234,
            mitre_tactics=[],
            mitre_techniques=[],
            settings=DetectorSettings([SensitivityMeterSetting(50)]),
            exclusion_settings=[],  # type: ignore
            run_status=DetectorRunStatus(DetectorStatusLabel.NOT_STARTED),
        )

    def _get_mock_s3_src_score(self) -> pd.DataFrame:
        s3_src_scores = [
            ["vmA", 3, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ["vmB", 3, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ["vmC", 3, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
        ]
        s3_src = pd.DataFrame(s3_src_scores, columns=["srcVmId", "cardinalityScore", "timestamp"])
        s3_src["timestamp"] = pd.to_datetime(s3_src["timestamp"], format="%Y-%m-%d %H:%M:%S").astype(str)
        s3_src.set_index("srcVmId", inplace=True)

        return s3_src

    def _get_mock_s3_dst_score(self) -> pd.DataFrame:
        s3_dst_scores = [
            ["vmD", 3, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ["vmE", 3, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ["vmF", 3, datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
        ]
        s3_dst = pd.DataFrame(s3_dst_scores, columns=["dstVmId", "cardinalityScore", "timestamp"])
        s3_dst["timestamp"] = pd.to_datetime(s3_dst["timestamp"], format="%Y-%m-%d %H:%M:%S").astype(str)
        s3_dst.set_index("dstVmId", inplace=True)

        return s3_dst

    def test_summary(self):
        #  | vm | dstFwAllowed | extDstAllowed | srcFwAllowed | extSrcAllowed | dstFwBlocked | srcFwBlocked
        #  | combinedDstAllowed | combinedSrcAllowed | totalAllowed | totalBlocked | allowRatio | blockRatio | total
        #  | timestamp

        src_baselines = [
            [
                "vmA",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc),
            ],
            [
                "vmB",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc),
            ],
            [
                "vmC",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc),
            ],
            [
                "vmA",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc),
            ],
            [
                "vmB",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc),
            ],
            [
                "vmC",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc),
            ],
            [
                "vmA",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc),
            ],
            [
                "vmB",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc),
            ],
            [
                "vmC",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc),
            ],
            [
                "vmA",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc),
            ],
            [
                "vmB",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc),
            ],
            [
                "vmC",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc),
            ],
            [
                "vmA",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc),
            ],
            [
                "vmB",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc),
            ],
            [
                "vmC",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc),
            ],
        ]

        src_baselines = pd.DataFrame(
            src_baselines,
            columns=[
                "srcVmId",
                "dstFwAllowed",
                "extDstAllowed",
                "srcFwAllowed",
                "extSrcAllowed",
                "dstFwBlocked",
                "srcFwBlocked",
                "combinedDstAllowed",
                "combinedSrcAllowed",
                "totalAllowed",
                "totalBlocked",
                "allowRatio",
                "blockRatio",
                "total",
                "timestamp",
            ],
        )

        src_baselines.set_index("srcVmId", inplace=True)

        src_baselines["timestamp"] = pd.to_datetime(src_baselines["timestamp"], format="%Y-%m-%d %H:%M:%S").astype(str)

        src_today = [
            [
                "vmA",
                2,
                3,
                3,
                2,
                90,
                60,
                5,
                5,
                10,
                150,
                0.08125,
                0.9375,
                160,
                datetime(year=2023, month=1, day=6, tzinfo=timezone.utc),
            ],
            [
                "vmB",
                13,
                7,
                10,
                20,
                125,
                75,
                20,
                30,
                50,
                200,
                0.052,
                0.8,
                250,
                datetime(year=2023, month=1, day=6, tzinfo=timezone.utc),
            ],
            [
                "vmC",
                7,
                13,
                15,
                15,
                200,
                150,
                20,
                30,
                50,
                350,
                0.0325,
                0.875,
                400,
                datetime(year=2023, month=1, day=6, tzinfo=timezone.utc),
            ],
        ]

        src_today = pd.DataFrame(
            src_today,
            columns=[
                "srcVmId",
                "dstFwAllowed",
                "extDstAllowed",
                "srcFwAllowed",
                "extSrcAllowed",
                "dstFwBlocked",
                "srcFwBlocked",
                "combinedDstAllowed",
                "combinedSrcAllowed",
                "totalAllowed",
                "totalBlocked",
                "allowRatio",
                "blockRatio",
                "total",
                "timestamp",
            ],
        )

        src_today.set_index("srcVmId", inplace=True)

        dst_baselines = [
            [
                "vmD",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc),
            ],
            [
                "vmE",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc),
            ],
            [
                "vmF",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc),
            ],
            [
                "vmD",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc),
            ],
            [
                "vmE",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc),
            ],
            [
                "vmF",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc),
            ],
            [
                "vmD",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc),
            ],
            [
                "vmE",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc),
            ],
            [
                "vmF",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc),
            ],
            [
                "vmD",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc),
            ],
            [
                "vmE",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc),
            ],
            [
                "vmF",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc),
            ],
            [
                "vmD",
                1,
                2,
                6,
                4,
                5,
                8,
                3,
                10,
                13,
                13,
                0.5,
                0.5,
                26,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc),
            ],
            [
                "vmE",
                2,
                3,
                5,
                3,
                5,
                7,
                5,
                8,
                13,
                12,
                0.52,
                0.48,
                25,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc),
            ],
            [
                "vmF",
                3,
                1,
                3,
                4,
                9,
                10,
                14,
                7,
                21,
                19,
                0.325,
                0.475,
                40,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc),
            ],
        ]

        dst_baselines = pd.DataFrame(
            dst_baselines,
            columns=[
                "dstVmId",
                "dstFwAllowed",
                "extDstAllowed",
                "srcFwAllowed",
                "extSrcAllowed",
                "dstFwBlocked",
                "srcFwBlocked",
                "combinedDstAllowed",
                "combinedSrcAllowed",
                "totalAllowed",
                "totalBlocked",
                "allowRatio",
                "blockRatio",
                "total",
                "timestamp",
            ],
        )

        dst_baselines.set_index("dstVmId", inplace=True)

        dst_baselines["timestamp"] = pd.to_datetime(dst_baselines["timestamp"], format="%Y-%m-%d %H:%M:%S").astype(str)

        dst_today = [
            [
                "vmD",
                2,
                3,
                3,
                2,
                90,
                60,
                5,
                5,
                10,
                150,
                0.08125,
                0.9375,
                160,
                datetime(year=2023, month=1, day=6, tzinfo=timezone.utc),
            ],
            [
                "vmE",
                13,
                7,
                10,
                20,
                125,
                75,
                20,
                30,
                50,
                200,
                0.052,
                0.8,
                250,
                datetime(year=2023, month=1, day=6, tzinfo=timezone.utc),
            ],
            [
                "vmF",
                7,
                13,
                15,
                15,
                200,
                150,
                20,
                30,
                50,
                350,
                0.0325,
                0.875,
                400,
                datetime(year=2023, month=1, day=6, tzinfo=timezone.utc),
            ],
        ]

        dst_today = pd.DataFrame(
            dst_today,
            columns=[
                "dstVmId",
                "dstFwAllowed",
                "extDstAllowed",
                "srcFwAllowed",
                "extSrcAllowed",
                "dstFwBlocked",
                "srcFwBlocked",
                "combinedDstAllowed",
                "combinedSrcAllowed",
                "totalAllowed",
                "totalBlocked",
                "allowRatio",
                "blockRatio",
                "total",
                "timestamp",
            ],
        )

        dst_today.set_index("dstVmId", inplace=True)

        # derived df from baseline
        src_derived_baseline = Baseline.transform_combined_baseline(src_baselines, "site1", TrafficDirection.SRC)
        dst_derived_baseline = Baseline.transform_combined_baseline(dst_baselines, "site1", TrafficDirection.DST)

        print(src_derived_baseline)
        print(dst_derived_baseline)

        # stage 1
        """
        | srcVmId                              |   sigmaStdDev |   stdDevScore |
        |:-------------------------------------|--------------:|--------------:|
        | 44783bbf-cc0a-4194-8df4-71d5a6d1f798 |       401.443 |           150 |
        """
        s1_src = self.detector._execute_blocked_count_std_stage(
            src_derived_baseline, src_today, "site1", TrafficDirection.SRC
        )
        s1_dst = self.detector._execute_blocked_count_std_stage(
            dst_derived_baseline, dst_today, "site1", TrafficDirection.DST
        )

        print(s1_src)
        print(s1_dst)

        # stage 2
        """
        | srcVmId                              |   trendChange |   trendChangeDiff |   trendChangeNormFactor |   totalMean |
        |:-------------------------------------|--------------:|------------------:|------------------------:|------------:|
        | 44783bbf-cc0a-4194-8df4-71d5a6d1f798 |           100 |           37454.7 |                 375.947 |     5.35714 |
        """
        s2_src = self.detector._execute_trend_change_stage(
            src_derived_baseline, src_today, "site1", TrafficDirection.SRC
        )
        s2_dst = self.detector._execute_trend_change_stage(
            dst_derived_baseline, dst_today, "site1", TrafficDirection.DST
        )

        print(s2_src)
        print(s2_dst)

        # TODO: stage 3: not called since it uses druid query
        # s3_src = self.detector._execute_port_cardinality_stage(s2_src, "site1", TrafficDirection.SRC)
        # s3_dst = self.detector._execute_port_cardinality_stage(s2_dst, "site1", TrafficDirection.DST)

        # Fake out stage 3 since it uses druid
        """
        | srcVmId                              |   cardinalityCount | timestamp                |
        |:-------------------------------------|-------------------:|:-------------------------|
        | 44783bbf-cc0a-4194-8df4-71d5a6d1f798 |             3.0022 | 2023-04-11T00:00:00.000Z |
        """
        s3_src = self._get_mock_s3_src_score()
        s3_dst = self._get_mock_s3_dst_score()

        # Combine all score
        f_src_score = self.detector._combine_score(
            s1_src, s2_src, s3_src, self._getDefaultUserConfig(), "site1", TrafficDirection.SRC
        )

        f_dst_score = self.detector._combine_score(
            s1_dst, s2_dst, s3_dst, self._getDefaultUserConfig(), "site1", TrafficDirection.DST
        )

        print(f_src_score)
        print(f_dst_score)

        """
        vmA:
        stdDevScore                        101.520256
        trendChange                             100.0
        cardinalityScore                            3
        timestamp           2023-01-01 00:00:00+00:00
        Score                               76.510128
        is_silent                               False
        Name: vmA, dtype: object

        vmB:
        stdDevScore                        104.201026
        trendChange                             100.0
        cardinalityScore                            3
        timestamp           2023-01-01 00:00:00+00:00
        Score                               77.850513
        is_silent                               False
        Name: vmB, dtype: object

        vmC:
        stdDevScore                        118.831026
        trendChange                             100.0
        cardinalityScore                            3
        timestamp           2023-01-01 00:00:00+00:00
        Score                               85.165513
        is_silent                               False
        """

        self.assertEqual(f_src_score.index.isin(["vmA", "vmB", "vmC"]).all(), True)
        self.assertEqual(f_dst_score.index.isin(["vmD", "vmE", "vmF"]).all(), True)

        self.assert_(f_src_score.loc["vmA"]["stdDevScore"] > 100)
        self.assert_(f_src_score.loc["vmB"]["stdDevScore"] > 100)
        self.assert_(f_src_score.loc["vmC"]["stdDevScore"] > 100)

        self.assert_(f_src_score.loc["vmA"]["Score"] > 75)
        self.assert_(f_src_score.loc["vmB"]["Score"] > 75)
        self.assert_(f_src_score.loc["vmC"]["Score"] > 75)

        self.assert_(f_dst_score.loc["vmD"]["stdDevScore"] > 100)
        self.assert_(f_dst_score.loc["vmE"]["stdDevScore"] > 100)
        self.assert_(f_dst_score.loc["vmF"]["stdDevScore"] > 100)

        self.assert_(f_dst_score.loc["vmD"]["Score"] > 75)
        self.assert_(f_dst_score.loc["vmE"]["Score"] > 75)
        self.assert_(f_dst_score.loc["vmF"]["Score"] > 75)

    def test_get_query_intervals(self):
        result = self.detector._get_baseline_intervals([])
        self.assertEqual([], result)

        today = datetime.combine(date.today(), datetime.min.time())
        result = self.detector._get_baseline_intervals([Interval(today - timedelta(days=7), today)])

        self.assertEqual(7, len(result))

    def test_build_event(self):
        today = datetime.combine(date.today(), datetime.min.time())
        counts = [
            [1, today],
            [32, today - timedelta(days=1)],
            [42, today - timedelta(days=2)],
            [62, today - timedelta(days=3)],
            [33, today - timedelta(days=4)],
            [9, today - timedelta(days=5)],
        ]
        blocked_counts_df = pd.DataFrame(counts, columns=["totalBlocked", "timestamp"])
        allowed_counts_df = pd.DataFrame(counts, columns=["totalAllowed", "timestamp"])

        event_info = (
            Builder()
            .direction(TrafficDirection.SRC)
            .src_computes("vm1", [f"vm{i}" for i in range(2, 100)])
            .dst_computes("vm10", [f"vm1{i}" for i in range(1, 100)])
            .hot_spot(["vm1"])
            .config_changes(pd.DataFrame())
            .blocked_counts(blocked_counts_df)
            .allowed_counts(allowed_counts_df)
            .new_ports(["22", "443"])
            .description(TrafficDirection.SRC, "vm1")
            .viz_filter()
            .compute_ip_map({"vm1": "192.168.1.1"})
            .compute_id("vm1")
            .detector_type(DetectorType.NETWORK_TRAFFIC_DROP)
            .detection_time(int(today.timestamp()))
            .confidence(50)
            .silent(False)
            .site_id("site 1")
            .org_id("org 1")
            .build()
        )

        self.assertEqual(["site 1"], event_info.site_ids)
        self.assertEqual(Score(50, 20), event_info.score)

        self.assertEqual(
            1,
            len(event_info.event_metadata.kv_store.get("METADATA_KEY_SRC_COMPUTES").list_string_value),  # type: ignore
        )
        self.assertEqual(
            99,
            len(event_info.event_metadata.kv_store.get("METADATA_KEY_DST_COMPUTES").list_string_value),  # type: ignore
        )
        self.assertEqual(
            1,
            len(event_info.event_metadata.kv_store.get("METADATA_KEY_HOTSPOTS").list_string_value),  # type: ignore
        )

        expected_affected_compute = Compute(
            compute_type=ComputeType.COMPUTE_TYPE_VM,
            ip_address="192.168.1.1",
            compute_id="vm1",
            fqdn=None,
            site_id="site 1",
            ip_mask=IpMask.PRIVATE,
        )
        self.assertEqual(event_info.affected_compute, expected_affected_compute)

        self.assertTrue(event_info.is_client)
        self.assertTrue(event_info.is_attacker)

        expected_data_flows = [
            DataFlow(
                src_compute=expected_affected_compute,
                dst_compute=Compute(
                    compute_type=ComputeType.COMPUTE_TYPE_VM,
                    ip_address=None,
                    compute_id="vm11",
                    fqdn=None,
                    site_id="site 1",
                    ip_mask=IpMask.PRIVATE,
                ),
                transport_protocols=None,
                application_protocols=None,
                server_ports=None,
                start_time=int(today.timestamp()),
                end_time=int(today.timestamp()),
            ),
            DataFlow(
                src_compute=expected_affected_compute,
                dst_compute=Compute(
                    compute_type=ComputeType.COMPUTE_TYPE_VM,
                    ip_address=None,
                    compute_id="vm12",
                    fqdn=None,
                    site_id="site 1",
                    ip_mask=IpMask.PRIVATE,
                ),
                transport_protocols=None,
                application_protocols=None,
                server_ports=None,
                start_time=int(today.timestamp()),
                end_time=int(today.timestamp()),
            ),
        ]
        self.assertEqual(event_info.data_flows[:2], expected_data_flows)  # type: ignore

        # try again with limit applied
        original_compute_limit = Config.EVENT_METADATA_COMPUTE_LIMIT
        Config.EVENT_METADATA_COMPUTE_LIMIT = "2"
        event_info = (
            Builder()
            .direction(TrafficDirection.SRC)
            .src_computes("vm1", [f"vm{i}" for i in range(2, 100)])
            .dst_computes("vm10", [f"vm1{i}" for i in range(1, 100)])
            .hot_spot(["vm0", "vm1", "vm100"])
            .config_changes(pd.DataFrame())
            .blocked_counts(blocked_counts_df)
            .allowed_counts(allowed_counts_df)
            .new_ports(["22", "443"])
            .description(TrafficDirection.SRC, "vm1")
            .viz_filter()
            .compute_ip_map({"vm1": "192.168.1.1"})
            .compute_id("vm1")
            .detector_type(DetectorType.NETWORK_TRAFFIC_DROP)
            .detection_time(int(today.timestamp()))
            .confidence(50)
            .silent(False)
            .site_id("site 1")
            .org_id("org 1")
            .build()
        )

        self.assertEqual(
            ["vm1"],
            event_info.event_metadata.kv_store.get("METADATA_KEY_SRC_COMPUTES").list_string_value,  # type: ignore
        )
        self.assertEqual(
            ["vm11", "vm12"],
            event_info.event_metadata.kv_store.get("METADATA_KEY_DST_COMPUTES").list_string_value,  # type: ignore
        )
        self.assertEqual(
            ["vm0", "vm1"],
            event_info.event_metadata.kv_store.get("METADATA_KEY_HOTSPOTS").list_string_value,  # type: ignore
        )

        # reset compute limit for other tests
        Config.EVENT_METADATA_COMPUTE_LIMIT = original_compute_limit

    def test_create_event(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()

        compute_to_ip = {"vm1": "1.1.1.1"}

        self._mock_lookup_service.computes_to_ip_lookup = lambda: compute_to_ip

        compute_row = pd.Series([False, 50], index=["is_silent", "Score"])

        event_info = self.detector._create_event(
            "vm1", compute_row, "site 1", TrafficDirection.SRC, self._mock_lookup_service.computes_to_ip_lookup()
        )

        self.assertEqual(["site 1"], event_info.site_ids)
        self.assertEqual(DetectorType.NETWORK_TRAFFIC_DROP, event_info.detector_name)

    def test_get_detector_status(self):
        self.mock_baseline_service = Mock(BaselineService())
        self.mock_baseline_service.get_max_baseline_duration_days = lambda _x, _y, _z: 2
        original_baseline_service = self.detector._baseline_service
        self.detector._baseline_service = self.mock_baseline_service

        # Not enough baseline case
        detector_run_status = self.detector._get_detector_run_status(
            DetectorStatusLabel.SUCCESS, "123", 1234, "success"
        )
        self.assertEqual(DetectorStatusLabel.NOT_ENOUGH_BASELINE, detector_run_status.status_label)

        # FAILURE case
        detector_run_status = self.detector._get_detector_run_status(DetectorStatusLabel.FAILURE, "123", 1234, "failed")
        self.assertEqual(DetectorStatusLabel.FAILURE, detector_run_status.status_label)

        # SUCCESS case - baseline size > min required baseline
        self.mock_baseline_service.get_max_baseline_duration_days = lambda _x, _y, _z: 5
        detector_run_status = self.detector._get_detector_run_status(
            DetectorStatusLabel.SUCCESS, "123", 1234, "success"
        )
        self.assertEqual(DetectorStatusLabel.SUCCESS, detector_run_status.status_label)

        self.detector._baseline_service = original_baseline_service

    def test_get_excluded_compute_from_config(self):
        config = UserConfigInfo(
            site_id="site1",
            name="TDROP",
            org_id="org_1",
            version="1",
            last_modified_time=1234,
            enabled=True,
            silent=False,
            source="source",
            block=False,
            block_confidence_threshold=1234,
            mitre_tactics=[],
            mitre_techniques=[],
            settings=DetectorSettings([SensitivityMeterSetting(50)]),
            exclusion_settings=AnomalyExclusionSettings(
                [
                    AnomalyExclusionSetting(
                        ExclusionLabel.DESTINATION, ExcludedItemsInfo(["vm1"]), ExclusionTypeInfo.VM
                    ),
                    AnomalyExclusionSetting(
                        ExclusionLabel.DESTINATION, ExcludedItemsInfo(["grp1", "grp2"]), ExclusionTypeInfo.GROUP
                    ),
                ]
            ),
            run_status=DetectorRunStatus(DetectorStatusLabel.NOT_STARTED),
        )
        excluded_computes = ["vm1", "grp1", "grp2"]

        self.mock_lookup_service = Mock(LookupService(Mock(FlowService)))
        self.mock_lookup_service.groups_to_computes_lookup = lambda _x, _y: set(["grp1", "grp2"])
        original_lookup_service = self.detector._lookup_service
        self.detector._lookup_service = self.mock_lookup_service
        self.assertEquals(set(excluded_computes), set(self.detector._get_excluded_computes_from_config(config)))
        self.detector._lookup_service = original_lookup_service
