import unittest
from datetime import date, datetime, timedelta, timezone
from unittest.mock import Mock

import pandas as pd
from common.utils.common_model import DetectorType, Interval
from common.utils.flow.core import FlowService
from trafficdrop.detector.common_types import TrafficDirection
from trafficdrop.detector.query import DruidQuery


class TestQuery(unittest.TestCase):
    def setUp(self):
        self._mock_flow_service = Mock(FlowService)
        self.druid_query = DruidQuery(
            DetectorType.NETWORK_TRAFFIC_DROP.name,
            self._mock_flow_service
        )

    def test_execute_base_traffic_drop_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()

        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_base_traffic_drop_query(
            [Interval(today - timedelta(days=1), today)],
            "site 1",
            TrafficDirection.SRC,
            [],
            []
        )

        self.assertEqual(0, len(result.index))

        src_baselines = [
            ["vmA", 1, 2, 6, 4, 5, 8, 3, 10, 13, 13, 0.5, 0.5, 26,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ["vmB", 2, 3, 5, 3,	5, 7, 5, 8,	13,	12,	0.52, 0.48, 25,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],
            ["vmC", 3, 1, 3, 4, 9, 10, 14, 7, 21, 19, 0.325, 0.475, 40,
                datetime(year=2023, month=1, day=1, tzinfo=timezone.utc)],

            ["vmA", 1, 2, 6, 4, 5, 8, 3, 10, 13, 13, 0.5, 0.5, 26,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
            ["vmB", 2, 3, 5, 3,	5, 7, 5, 8,	13,	12,	0.52, 0.48, 25,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],
            ["vmC", 3, 1, 3, 4, 9, 10, 14, 7, 21, 19, 0.325, 0.475, 40,
                datetime(year=2023, month=1, day=2, tzinfo=timezone.utc)],

            ["vmA", 1, 2, 6, 4, 5, 8, 3, 10, 13, 13, 0.5, 0.5, 26,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
            ["vmB", 2, 3, 5, 3,	5, 7, 5, 8,	13,	12,	0.52, 0.48, 25,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],
            ["vmC", 3, 1, 3, 4, 9, 10, 14, 7, 21, 19, 0.325, 0.475, 40,
                datetime(year=2023, month=1, day=3, tzinfo=timezone.utc)],

            ["vmA", 1, 2, 6, 4, 5, 8, 3, 10, 13, 13, 0.5, 0.5, 26,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
            ["vmB", 2, 3, 5, 3,	5, 7, 5, 8,	13,	12,	0.52, 0.48, 25,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],
            ["vmC", 3, 1, 3, 4, 9, 10, 14, 7, 21, 19, 0.325, 0.475, 40,
                datetime(year=2023, month=1, day=4, tzinfo=timezone.utc)],

            ["vmA", 1, 2, 6, 4, 5, 8, 3, 10, 13, 13, 0.5, 0.5, 26,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
            ["vmB", 2, 3, 5, 3,	5, 7, 5, 8,	13,	12,	0.52, 0.48, 25,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
            ["vmC", 3, 1, 3, 4, 9, 10, 14, 7, 21, 19, 0.325, 0.475, 40,
                datetime(year=2023, month=1, day=5, tzinfo=timezone.utc)],
        ]

        src_baselines = pd.DataFrame(
            src_baselines,
            columns=["srcVmId", "dstFwAllowed", "extDstAllowed", "srcFwAllowed", "extSrcAllowed", "dstFwBlocked",
                     "srcFwBlocked", "combinedDstAllowed", "combinedSrcAllowed", "totalAllowed", "totalBlocked",
                     "allowRatio", "blockRatio", "total", "timestamp"])

        self._mock_flow_service.group_by = lambda _: src_baselines

        result = self.druid_query.execute_base_traffic_drop_query(
            [Interval(today - timedelta(days=1), today)],
            "site 1",
            TrafficDirection.SRC,
            ["vm1", "vm2"],
            ["vm3"]
        )

        self.assertEqual(15, len(result.index))

    def test_execute_port_cardinality_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()
        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_port_cardinality_query(
            [Interval(today - timedelta(days=1), today)],
            ["vm1", "vm2", "vm3"],
            "site 1",
            TrafficDirection.SRC
        )

        self.assertEqual(0, len(result.index))

        mock_query_result = [
            ["vm1", 12, datetime(year=2023, month=1, day=5,
                                 tzinfo=timezone.utc)],
            ["vm2", 34, datetime(year=2023, month=1,
                                 day=5, tzinfo=timezone.utc)]
        ]

        mock_query_df = pd.DataFrame(mock_query_result, columns=[
                                     "srcVmId", "cardinalityCount", "timestamp"])

        self._mock_flow_service.group_by = lambda _: mock_query_df

        result = self.druid_query.execute_port_cardinality_query(
            [Interval(today - timedelta(days=1), today)],
            ["vm1", "vm2", "vm3"],
            "site 1",
            TrafficDirection.SRC
        )

        self.assertEqual(2, len(result.index))

    def test_execute_rule_id_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()
        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_rule_id_query(
            [Interval(today - timedelta(days=1), today)],
            "vm1",
            "site 1",
            TrafficDirection.SRC
        )

        self.assertEqual(0, len(result.index))

    def test_execute_get_policy_path_for_rule_id(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()
        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_get_policy_path_for_rule_id(
            Interval(today - timedelta(days=1), today),
            "site 1",
            TrafficDirection.SRC,
            ["rule 1", "rule 2"]
        )

        self.assertEqual(0, len(result.index))

    def test_execute_fw_event_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()
        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_fw_event_query(
            [Interval(today - timedelta(days=1), today)],
            "site 1",
            TrafficDirection.SRC
        )
        self.assertEqual(0, len(result.index))

    def test_execute_get_dstport_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()
        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_get_dstport_query(
            [Interval(today - timedelta(days=1), today)],
            "vm1",
            "site 1",
            TrafficDirection.SRC
        )
        self.assertEqual(0, len(result.index))

    def test_execute_opposite_compute_query(self):
        self._mock_flow_service.group_by = lambda _: pd.DataFrame()
        today = datetime.combine(date.today(), datetime.min.time())
        result = self.druid_query.execute_opposite_compute_query(
            Interval(today - timedelta(days=1), today),
            "vm1",
            "site 1",
            TrafficDirection.SRC
        )
        self.assertEqual(0, len(result.index))
